"""Gathers the formulas for the solubility of NaCl solutions.

Sources
-------
- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

(some info of domain of validity of expressions in the following paper:)
Dorn, J. & Steiger, M. Measurement and Calculation of Solubilities in the
Ternary System NaCH 3 COO + NaCl + H 2 O from 278 K to 323 K.
J. Chem. Eng. Data 52, 1784-1790 (2007).)

- CRC Handbook of Chemistry and Physics, 104th Edition
  Section: 4 | Solubility of Common Inorganic Salts as a Function of Temperature
  https://hbcp.chemnetbase.com/documents/04_29/04_29_0001.xhtml?dswid=7662

- Sparrow, B. S.
  Empirical equations for the thermodynamic properties of aqueous sodium chloride. Desalination 159, 161-170
  (2003).
"""

from ...general import SaturatedSolutionFormula
from .steiger import Solubility_NaCl_Steiger2008_Base
from .steiger import Solubility_NaCl_2H2O_Steiger2008_Base
from .crc_handbook import Solubility_NaCl_CRCHandbook_Base


class Solubility_NaCl_Steiger(Solubility_NaCl_Steiger2008_Base):
    """Already defined in steiger module; default for NaCl anhydrous"""
    default = True


class Solubility_NaCl_2H2O_Steiger(Solubility_NaCl_2H2O_Steiger2008_Base):
    """Already defined in steiger module; default for hydrohalite"""
    default = True


class Solubility_NaCl_CRCHandbook(Solubility_NaCl_CRCHandbook_Base):
    """Already defined in CRC Handbook module"""
    pass


class Solubility_NaCl_Sparrow(SaturatedSolutionFormula):

    source = 'Sparrow'
    crystal = 'NaCl'

    temperature_unit = 'C'
    temperature_range = (0, 450)

    concentration_unit = 'w'

    coeffs = 0.2628, 62.75e-6, 1.084e-6

    def calculate(self, T):
        a0, a1, a2 = self.coeffs
        return a0 + a1 * T + a2 * T**2


# ============================= WRAP-UP FORMULAS =============================

SolubilityFormulas_NaCl = (
    Solubility_NaCl_Steiger,
    Solubility_NaCl_2H2O_Steiger,
    Solubility_NaCl_CRCHandbook,
    Solubility_NaCl_Sparrow
)


"""Implementation of Pitzer equations (activity, volumetric).

Mostly based on Michael Steiger's group papers.
"""

import numpy as np

from ...constants import R as ideal_gas_constant
from ...constants import Mw, charge_numbers, dissociation_numbers
from .ionic import ionic_strength


b = 1.2  # kg^1/2.mol^-1/2


class PitzerBase:
    """Base class for Pitzer equations for water & solute activity, or volumetric."""

    def __init__(self, T=298.15, solute='NaCl', R=None, **coeffs):
        """Init PitzerBase object.

        Input
        -----
        T: temperature in K
        solute: solute name (str, default 'NaCl')
        R: (optional) replacement value for the ideal gas constant, R
           (useful if e.g. units use atmospheres instead of MPa etc.)
        coeffs: Pitzer coefficients
        """
        self.T = T
        self.R = R if R is not None else ideal_gas_constant
        self.solute = solute

        for coeff_name, coeff_value in coeffs.items():
            setattr(self, coeff_name, coeff_value)

        self.z_m, self.z_x = charge_numbers[solute]
        self.nu_m, self.nu_x = dissociation_numbers[solute]
        self.nu_mx = self.nu_m + self.nu_x

    @staticmethod
    def g(x):
        """Function used to calculate the second virial coefficients"""
        y = 2 * (1 - (1 + x) * np.exp(-x))
        # Below is a trick to avoid the problem of division by zero, since g(0) = 0
        return np.divide(y, x**2, out=np.ones_like(x), where=x != 0)

    def virial_B(self, m, func=None):
        """Calculate second virial coefficient at a given molality

        Input
        -----
        - m: molality in mol/kg
        - func: which function used with beta1 / beta2 (default: g(x))

        Output
        ------
        Second Virial coefficients in units corresponding to input coeffs.
        """
        func = self.g if func is None else func
        sqrtI = ionic_strength(self.solute, m=m)**(1 / 2)
        x1 = self.alpha1 * sqrtI
        x2 = self.alpha2 * sqrtI
        return self.beta0 + self.beta1 * func(x1) + self.beta2 * func(x2)


class PitzerVolumetric(PitzerBase):
    """Implementation of Pitzer Equations for volumetric properties"""

    def __init__(self, T=298.15, solute='NaCl', R=None, **coeffs):
        """Init PitzerVolumertic object.

        Input
        -----
        T: temperature in K
        solute: solute name (str, default 'NaCl')
        R: (optional) replacement value for the ideal gas constant, R
           (useful if e.g. units use atmospheres instead of MPa etc.)
        coeffs: must contain the following keys:
        - A_v [cm3·kg1/2·mol-3/2]: first virial coefficient
        - beta0, beta1 [kg·mol-1·MPa-1]: second virial coefficient parameters
        - C_v [kg2·mol-2·MPa-1]: third virial coefficient
        - v_0 [cm3/mol]: apparent molar volume at infinite dilution

        optional keys (to specify if electrolyte is 2:2)
        - beta2 (default 0)
        - alpha1 [kg1/2mol-1/2] (default 2)
        - alpha2 [kg1/2mol-1/2] (default 0)

        NOTE: the beta coefficients are in [kg·mol-1·MPa-1]
        CAUTION: non-SI unit (makes output of molar volumes in cm^3 / mol)

        NOTE: second Virial coefficients Bv are in [kg·mol-1·MPa-1].
        """
        coeffs['beta2'] = coeffs.get('beta2', 0)    # add default values
        coeffs['alpha1'] = coeffs.get('alpha1', 2)  # (these alpha values do not work for 2:2 electrolytes)
        coeffs['alpha2'] = coeffs.get('alpha2', 0)
        super().__init__(T=T, solute=solute, R=R, **coeffs)

    def apparent_molar_volume(self, m):
        """Apparent molar volume in cm^3 / mol at given molality

         Input
        -----
        - m: molality in mol/kg

        Output
        ------
        Apparent molar volume in cm^3 / mol (NON-SI)
        """
        Av = self.A_v
        Bv = self.virial_B(m=m)
        Cv = self.C_v
        T = self.T
        R = self.R

        sqrtI = ionic_strength(self.solute, m=m)**(1 / 2)

        v0 = self.v_0
        v1 = self.nu_mx * self.z_m * self.z_x * Av / (2 * b) * np.log(1 + b * sqrtI)
        v2 = 2 * R * T * self.nu_m * self.nu_x * (Bv * m + (self.nu_m * self.z_m * Cv) * m**2)

        return v0 + v1 + v2


class PitzerActivity(PitzerBase):
    """Implementation of Pitzer equations for activity (phi / gamma)"""

    def __init__(self, T=298.15, solute='NaCl', **coeffs):
        """Init PitzerActivity object.

        Input
        -----
        T: temperature in K
        solute: solute name (str, default 'NaCl')
        coeffs: must contain the following keys:
        - A_phi: first virial coefficient
        - alpha1, alpha2, beta0, beta1, beta2: 2d virial coefficient parameters
        - C_phi: third virial coefficient
        """
        super().__init__(T=T, solute=solute, **coeffs)
        self.C_gamma = 3 / 2 * self.C_phi

    def _activity_coeff(self, m, f, B, C):
        """Generic calculation for both osmotic coeff and ln(gamma).

        Input:
        m: molality
        f: Debye factor (f_phi or f_gamma)
        B, C: second and third virial coefficients
        """
        p1 = self.z_m * self.z_x * f
        p2 = B * 2 * m * (self.nu_m * self.nu_x) / self.nu_mx
        p3 = C * 2 * m**2 * (self.nu_m * self.nu_x)**(3 / 2) / self.nu_mx
        return p1 + p2 + p3

    @staticmethod
    def _g_phi(x):
        """Replacement function for B_phi calculation from virial_B"""
        return np.exp(-x)

    def B_phi(self, m):
        """Second virial coefficient for activity"""
        return self.virial_B(m, func=self._g_phi)

    def B_gamma(self, m):
        """Second virial coefficient for activity"""
        return self.B_phi(m) + self.virial_B(m, func=self.g)

    def f_phi(self, m):
        """Debye factor for osmotic coefficient"""
        sqrtI = ionic_strength(self.solute, m=m)**(1 / 2)
        return -self.A_phi * sqrtI / (1 + b * sqrtI)

    def f_gamma(self, m):
        """Debye factor for activity coefficient"""
        sqrtI = ionic_strength(self.solute, m=m)**(1 / 2)
        a1 = sqrtI / (1 + b * sqrtI)
        a2 = (2 / b) * np.log(1 + b * sqrtI)
        return -self.A_phi * (a1 + a2)

    def osmotic_coefficient(self, m):
        """Osmotic coefficient, phi"""
        coeff = self._activity_coeff(
            m,
            f=self.f_phi(m),
            B=self.B_phi(m),
            C=self.C_phi,
        )
        return 1 + coeff

    def water_activity(self, m):
        """Water activity, aw"""
        phi = self.osmotic_coefficient(m)
        return np.exp(-phi * Mw * self.nu_mx * m)

    def activity_coefficient(self, m):
        """Activity coefficient, gamma"""
        ln_gamma = self._activity_coeff(
            m,
            f=self.f_gamma(m),
            B=self.B_gamma(m),
            C=self.C_gamma,
        )
        return np.exp(ln_gamma)

    def solute_activity(self, m):
        """The complicated prefactor is due to how gamma_+/- etc. are defined

        (cf e.g. Steiger 2005)
        """
        prefactor = (self.nu_m ** self.nu_m) * (self.nu_x ** self.nu_x)
        gamma = self.activity_coefficient(m=m)
        return  prefactor * (gamma * m) ** self.nu_mx


class PitzerActivityOriginal(PitzerActivity):
    """In the original Pitzer 1973 paper, there are no alpha2, beta2 terms.

    (but just alpha=alpha1=2 fixed, A_phi = 0.392 fixed
    and beta0, beta1, C_phi tabulated)
    """
    def __init__(self, T=298.15, solute='NaCl', **coeffs):
        """Init PitzerActivity object.

        Input
        -----
        T: temperature in K
        solute: solute name (str, default 'NaCl')
        coeffs: must contain the following keys:
        - beta0, beta1, 2d virial coefficient parameters
        - C_phi: third virial coefficient
        """
        coeffs['A_phi'] = 0.392
        coeffs['alpha1'] = 2
        coeffs['alpha2'] = 0
        coeffs['beta2'] = 0
        super().__init__(T=T, solute=solute, **coeffs)



"""Tests for the aquasol.solutions module."""

import numpy as np

from aquasol.solutions import activity_coefficient
from aquasol.solutions import water_activity
from aquasol.solutions import density
from aquasol.solutions import surface_tension
from aquasol.solutions import refractive_index
from aquasol.solutions import electrical_conductivity
from aquasol.solutions import solubility, aw_saturated

from aquasol.solutions import osmotic_coefficient, osmotic_pressure
from aquasol.solutions import aw_to_conc
from aquasol.solutions import convert

from aquasol.constants import molar_mass
from aquasol.constants import charge_numbers
from aquasol.constants import dissociation_numbers


# ============================== Test constants ==============================


def test_constants():
    solute = 'Na2SO4'
    assert round(molar_mass(solute), 3) == 0.142  # kg / mol
    assert charge_numbers[solute] == (1, 2)
    assert dissociation_numbers[solute] == (2, 1)


# # =========================== Test activity coeff ============================


def test_gamma_1():
    gamma = activity_coefficient(m=[6.1, 10])  # around saturation for NaCl
    assert round(gamma[0], 2) == 1.0

def test_gamma_2():
    gamma = activity_coefficient(m=0)  # at infinite dilution
    assert round(gamma, 2) == 1.0

def test_gamma_3():
    gamma = activity_coefficient(m=0, solute='KCl')  # at infinite dilution
    assert round(gamma, 2) == 1.0

def test_gamma_KCl():
    kwargs = {'c': 2000, 'solute': 'KCl'}
    g1 = activity_coefficient(source='Tang', **kwargs)
    g2 = activity_coefficient(source='Steiger 2008', **kwargs)
    assert (round(g1, 2) == round(g2, 2) == 0.57)

def test_gamma_KCl_supersaturated():
    g1 = activity_coefficient('KCl', T=5, m=6)
    g2 = activity_coefficient('KCl', T=45, m=6)
    assert round(np.log(g1), 2) == -0.56
    assert round(np.log(g2), 2) == -0.46

def test_gamma_Na2SO4():
    kwargs = {'m': 10, 'solute': 'Na2SO4'}
    g1 = activity_coefficient(source='Steiger 2005', **kwargs)
    g2 = activity_coefficient(source='Steiger 2008', **kwargs)
    assert round(g1, 2) == round(g2, 2) == 0.16

def test_gamma_NaCl():
    kwargs = {'x': 0.15, 'solute': 'NaCl'}
    g1 = activity_coefficient(source='Steiger 2005', **kwargs)
    g2 = activity_coefficient(source='Steiger 2008', **kwargs)
    g3 = activity_coefficient(source='Tang', **kwargs)
    assert (round(g1, 1) == round(g2, 1) == round(g3, 1) == 1.5)


# =========================== Test water activity ============================


def test_aw_1():
    aw = water_activity(x=0.1)  # a_w for a mole fraction of 0.1 of NaCl
    assert round(aw, 2) == 0.75

def test_aw_2():
    aw = water_activity(w=0.2)  # a_w for a mass fraction of 0.2 of NaCl
    assert round(aw, 2) == 0.84

def test_aw_3():
    aw = water_activity(c=5000)  # a_w for a molality of 5 mol/L of NaCl
    assert round(aw, 2) == 0.78

def test_aw_4():
    aw = water_activity(m=6)      # a_w for a molality of 6 mol/kg of NaCl
    assert round(aw, 2) == 0.76

def test_aw_5():
    aw = water_activity('LiCl', m=6)  # same for LiCl
    assert round(aw, 2) == 0.68

def test_aw_6():
    aw = water_activity('LiCl', m=6, T=70)  # same for LiCl at 70°C
    assert round(aw, 2) == 0.70

def test_aw_7():
    aw = water_activity('LiCl', 293, 'K', m=6)  # same for LiCl at 293K.
    assert round(aw, 2) == 0.68

def test_aw_8():
    aw = water_activity(solute='CaCl2', T=50, m=[2, 4, 6])  # iterable conc.
    assert round(aw[2], 2) == 0.45

def test_aw_9():
    aw = water_activity(solute='KCl', m=3)  # KCl
    assert round(aw, 2) == 0.90

def test_aw_10():
    """Check different formulas are consistent"""
    aw1 = water_activity(solute='Na2SO4', m=4, source='Clegg')
    aw2 = water_activity(solute='Na2SO4', m=4, source='Steiger 2005')
    aw3 = water_activity(solute='Na2SO4', m=4, source='Steiger 2008')
    assert round(aw1, 2) == 0.85
    assert round(aw2, 2) == 0.85
    assert round(aw3, 2) == 0.85

def test_aw_11():
    """Check CaCl2"""
    aw = water_activity('CaCl2', w=0.1)  # CaCl2
    assert round(aw, 2) == 0.94


# Test extensions of water activity ------------------------------------------


def test_osmotic_pressure():
    pi = osmotic_pressure(m=4)
    assert round(pi / 1e6, 1) == 22.1


def test_osmotic_coefficient():
    phi = osmotic_coefficient(w=0.27)
    assert round(phi, 1) == 1.3


# =============================== Test density ===============================


def test_rho_1():
    rho = density(w=0.1)  # NaCl solution, at mass fraction of 0.1 at 25°C.
    assert round(rho) == 1069

def test_rho_2():
    rho = density('LiCl', 300, 'K', m=6)  # LiCl solution at 300K, 6 mol/kg
    assert round(rho) == 1116

def test_rho_3():
    rho = density(source='Tang', x=0.23)  # supersaturatad NaCl, Tang equation
    assert round(rho) == 1419

def test_rho_4():
    rho = density(c=5000, relative=True)  # relative density of NaCl,  5 mol/L.
    assert round(rho, 2) == 1.19

def test_rho_5():
    rho = density(w=[0.05, 0.12, 0.25])  # iterable concentration
    assert round(rho[2]) == 1186

def test_rho_Na2SO4():
    rho = density('Na2SO4', w=0.5)
    assert round(rho) == 1549

def test_rho_MgCl2():
    rho = density('MgCl2', w=0.1)
    assert round(rho) == 1082

def test_rho_KI():
    rho = density('KI', w=0.1)
    assert round(rho) == 1074

def test_rho_KCl():
    rho = density('KCl', w=0.1)
    assert round(rho) == 1061

def test_rho_CaCl2():
    rho = density('CaCl2', w=0.1)
    assert round(rho) == 1082


# =========================== Test surface tension ===========================


def test_sigma_1():
    s = surface_tension(x=0.09)  # NaCl at 25°C and a mole fraction of 9%
    assert round(s, 3) == 0.080

def test_sigma_2():
    s = surface_tension('LiCl', w=0.1)  # LiCl, 25°C and weight fract. of 10%
    assert round(s, 3) == 0.076

def test_sigma_3():
    s = surface_tension('CaCl2', 20, m=6.6666)  # CaCl2, 20°C, devil molality
    assert round(s, 3) == 0.096

def test_sigma_4():
    s = surface_tension('CaCl2', 353.15, 'K', c=5e3)  # CaCl2, 80°C, 5 mol/L
    assert round(s, 3) == 0.087

def test_sigma_5():
    s = surface_tension(x=[0.02, 0.04, 0.06, 0.08, 0.1], T=21)  # iterable conc.
    assert round(s[4], 3) == 0.082


# ========================== Test refractive index ===========================


def test_n_1():
    n = refractive_index(x=0.08)  # mole fraction of 0.08 of NaCl
    assert round(n, 2) == 1.37

def test_n_2():
    n = refractive_index(w=0.2)  # mass fraction of 0.2 of NaCl
    assert round(n, 2) == 1.37

def test_n_3():
    n = refractive_index(c=4321)  # molality of 4.321 mol/L of NaCl
    assert round(n, 2) == 1.37

def test_n_4():
    n = refractive_index(m=3)   # molality of 6 mol/kg of NaCl
    assert round(n, 2) == 1.36

def test_n_5():
    n = refractive_index('KCl', m=1.6)  # KCl, 1.6 mol/kg, 25°C
    assert round(n, 2) == 1.35

def test_n_6():
    n = refractive_index('KCl', m=1.9, T=40)  # KCl at 40°C, 1.9 mol/kg
    assert round(n, 2) == 1.35

def test_n_7():
    n = refractive_index('KCl', 312, 'K', m=1.9)  # KCl at 312K, 1.9 mol/kg
    assert round(n, 2) == 1.35

def test_n_8():
    n = refractive_index('KCl', T=22, w=[0.05, 0.1, 0.15])  # iterable conc.
    assert round(n[2], 2) == 1.36


# ====================== Test electrical conductivity ========================


def test_conduc_concs():
    s1, s2, s3 = electrical_conductivity('KCl', m=[0.01, 0.1, 1])  # At 25°C
    assert round(s1, 4) == 0.1408
    assert round(s2, 3) == 1.282
    assert round(s3, 2) == 10.86


def test_conduc_temps():
    s_0, s_25, s_50 = electrical_conductivity('KCl', m=1, T=[0, 25, 50])
    assert round(s_0, 2) == 6.35
    assert round(s_25, 2) == 10.86
    assert round(s_50, 2) == 15.75



# ============================= Test solubility ==============================


def test_solubility_1():
    m_sat = solubility()          # solubility (molality) of NaCl at 25°C
    assert round(m_sat, 2) == 6.15

def test_solubility_2():
    m_sat = solubility(T=40)      # solubility (molality) of NaCl at 40°C
    assert round(m_sat, 2) == 6.22

def test_solubility_3():
    x_sat = solubility(out='x')   # solubility (mole fraction) of NaCl at 25°C
    assert round(x_sat, 2) == 0.1

def test_solubility_4():
    c_sat = solubility(out='c')   # solubility (molarity) of NaCl at 25°C
    assert round(c_sat / 1000, 1) == 5.4

def test_solubility_6():
    m_sat = solubility(T=[10, 15, 20, 25, 30])     # iterables accepted too
    assert round(m_sat[-2], 2) == 6.15

def test_solubility_7():
    """Must correspond to CRC Handbook values"""
    m_sat_10 = solubility(T=5, source='CRC Handbook')
    m_sat_40 = solubility(T=40, source='CRC Handbook')
    assert round(m_sat_10, 2) == 6.11
    assert round(m_sat_40, 2) == 6.22

def test_solubility_7():
    """Must correspond to CRC Handbook values"""
    m_sat_10 = solubility('LiCl', T=10)
    m_sat_25 = solubility('LiCl', T=25)
    assert round(m_sat_10, 3) == 19.296
    assert round(m_sat_25, 3) == 19.935

def test_solubility_KCl():
    s1 = solubility('KCl', out='w', T=290, unit='K')
    s2 = solubility('KCl', out='m', T=45)
    assert round(s1, 2) == 0.25
    assert round(s2, 1) == 5.6

def test_solubility_Na2SO4():
    T = 32.38  # This is the point of equilibrium thenardite/mirabilite
    s1 = solubility('Na2SO4', out='m', T=T)
    s2 = solubility('Na2SO4,10H2O', out='m', T=T)
    assert round(s1, 2) == round(s2, 2) == 3.51

def test_solubility_Halite():
    T = 0.1  # This is the point of equilibrium halite/hydrohalite
    s1 = solubility('NaCl', out='m', T=T)
    s2 = solubility('NaCl,2H2O', out='m', T=T)
    assert round(s1, 2) == round(s2, 2) == 6.10

def test_solubility_LiBr():

    # solubiliity at 25°C
    assert round(solubility('LiBr,2H2O')) == 21

    # equilibrium dihydrate/trihydrate
    T = 5.7
    sa1 = solubility('LiBr,2H2O', T=T)
    sa2 = solubility('LiBr,3H2O', T=T)
    assert round(sa1, 1) == round(sa2, 1)

    # equilibrium dihydrate/monohydrate
    T = 34.6
    sb1 = solubility('LiBr,2H2O', T=T)
    sb2 = solubility('LiBr,H2O', T=T)
    assert round(sb1, 1) == round(sb2, 1)


# ------------------------- Extensions of solubility -------------------------

def test_aw_saturated_NaCl():
    aw_sat_25 = aw_saturated()     # NaCl at 25°C
    aw_sat_07 = aw_saturated(T=7)  # NaCl at 7°C
    assert round(100 * aw_sat_25, 1) == 75.3
    assert round(100 * aw_sat_07, 1) == 75.7

def test_aw_saturated_KCl():
    aw_sat_25 = aw_saturated('KCl')        # KCl at 25°C
    aw_sat_10 = aw_saturated('KCl', T=10)  # KCl at 10°C
    assert round(100 * aw_sat_25) == 84
    assert round(100 * aw_sat_10) == 87

def test_aw_saturated_4():
    aw_sat_25 = aw_saturated('LiCl')        # LiCl at 25°C
    aw_sat_10 = aw_saturated('LiCl', T=10)  # LiCl at 10°C
    assert round(100 * aw_sat_25) == 11
    assert round(100 * aw_sat_10) == 10

# =============================== Test convert ===============================

def test_convert_1():
    x = convert(0.4, 'w', 'x') # mass fraction 0.4 into mole fraction for NaCl
    assert round(x, 2) == 0.17

def test_convert_2():
    w = convert(10, 'm', 'w')  # molality 10 mol/kg into mass fraction of NaCl
    assert round(w, 2) == 0.37

def test_convert_3():
    w = convert(17, 'm', 'w', 'LiCl')  # 10 mol/kg of LiCl into mass fraction
    assert round(w, 2) == 0.42

def test_convert_4():
    x = convert(5120, 'c', 'x')  # 5.12 mol/m^3 to mole fraction, NaCl, 25°C
    assert round(x, 3) == 0.094

def test_convert_5():
    x = convert(3300, 'c', 'x', T=30)  # 3.3 mol/m^3 to x at 30°C
    assert round(x, 3) == 0.060

def test_convert_6():
    x = convert(3300, 'c', 'x', T=300, unit='K')  # same but at 300 K
    assert round(x, 3) == 0.060

def test_convert_7():
    x = convert(3300, 'c', 'x', solute='LiCl')  # LiCl at 25°C
    assert round(x, 3) == 0.060

def test_convert_8():
    x = convert([3300, 4400], 'c', 'x', solute='NaCl', T=40)  # iterable ok
    assert round(x[1], 3) == 0.081

def test_convert_9():
    m = convert(5305, 'c', 'm', density_source='Tang')  # different source
    assert round(m, 2) == 6


# # ========================= Test Inverse Functions ===========================


def test_ac_1():                # in terms of weight fracton
    w = aw_to_conc(0.45)
    assert round(w, 2) == 0.46

def test_ac_2():                # in terms of molality
    m = aw_to_conc([0.5, 0.75], out='m')
    assert round(m[1], 1) == 6.2

def test_ac_3():                # in terms of mass ratio, for LiCl, at 50°C
    r = aw_to_conc(0.11, 'r', 'LiCl', T=50)
    assert round(r, 2) == 0.92



"""Activity of solutions according to Pitzer original paper.

NOTE: Almost identical in structure to water_activity.pitzer

Sources
-------
Pitzer, K. S. & Mayorga, G.
Thermodynamics of electrolytes. II.
Activity and osmotic coefficients for strong electrolytes with one or both
ions univalent.
J. Phys. Chem. 77, 2300-2308
(1973)
"""

from ...general import SolutionFormula
from ..pitzer import PitzerActivityOriginal


class ActivityCoefficient_Pitzer_Base(SolutionFormula):

    source = 'Pitzer'

    temperature_unit = 'K'
    temperature_range = (298.15, 298.15)

    concentration_unit = 'm'

    with_water_reference = False

    def calculate(self, m, T):
        pitz = PitzerActivityOriginal(T=T, solute=self.solute, **self.coeffs)
        return pitz.activity_coefficient(m=m)


# ============================ Different solutes =============================


class ActivityCoefficient_NaCl_Pitzer_Base(ActivityCoefficient_Pitzer_Base):
    solute = 'NaCl'
    concentration_range = (0, 6)
    coeffs = {
        'beta0': 0.0765,
        'beta1': 0.2664,
        'C_phi': 0.00127,
    }


class ActivityCoefficient_LiCl_Pitzer_Base(ActivityCoefficient_Pitzer_Base):
    solute = 'LiCl'
    concentration_range = (0, 6)
    coeffs = {
        'beta0': 0.1494,
        'beta1': 0.3074,
        'C_phi': 0.00359,
    }


class ActivityCoefficient_LiBr_Pitzer_Base(ActivityCoefficient_Pitzer_Base):
    solute = 'LiBr'
    concentration_range = (0, 2.5)
    coeffs = {
        'beta0': 0.1748,
        'beta1': 0.2547,
        'C_phi': 0.0053,
    }



"""Water activity of solutions"""

from .CaCl2 import WaterActivityFormulas_CaCl2
from .KCl import WaterActivityFormulas_KCl
from .LiBr import WaterActivityFormulas_LiBr
from .LiCl import WaterActivityFormulas_LiCl
from .Na2SO4 import WaterActivityFormulas_Na2SO4
from .NaCl import WaterActivityFormulas_NaCl

WaterActivityFormulas = (
    WaterActivityFormulas_CaCl2 +
    WaterActivityFormulas_KCl +
    WaterActivityFormulas_LiBr +
    WaterActivityFormulas_LiCl +
    WaterActivityFormulas_Na2SO4 +
    WaterActivityFormulas_NaCl
)



"""Miscellaneous formatting tools for humidity

(NOT WITH format module due to circular import problems)
"""

import numpy as np

from .water import vapor_pressure


# ================================== Config ==================================

hparams = ['p', 'rh', 'aw']

hratio = {'rh': 1 / 100, 'aw': 1}  # factor to go from humidity to activity
msg_humidity_error = "Humidity argument can only be 'p=', 'rh=' or 'aw='"

# ============================================================================


def format_humidity(unit='C', T=25, source=None, out='p', **humidity):
    """Manage conversion between p=, rh= and aw= keywordsil.

    Parameters
    ----------
    - unit: temperature unit ('C' or 'K')
    - T: temperature, required only if rh or aw are used (optional for p)
    - source: literature source for the calculation (if None --> default)
    - out: output parameter ('p', 'rh' or 'aw')
    - humidity kwargs: can be 'rh=' (relative humidity in %), 'aw=' (vapor
    activity = rh / 100), 'p=' (partial water vapor pressure).

    Output
    ------
    p (partial vapor pressure in Pa, float), rh, or aw depending on 'out'.

    Note: cannot be in the aquasol.format module because it needs to import
    vapor_pressure, which causes circular imports problems.
    """
    try:
        hin, = humidity.keys()    # humidity keyword
        val, = humidity.values()  # check there is only one input humidity arg.
    except ValueError:
        raise KeyError(msg_humidity_error)

    if hin not in hparams:
        raise KeyError(msg_humidity_error)

    if out in hparams:
        hout = out
    else:
        raise ValueError(f'out parameter can only be in {hparams}')

    if hin == hout:
        return val

    elif 'p' in [hin, hout]:
        # need to convert to/from p to aw/rh --> need psat(T)
        if T is None:
            T = 25 if unit == 'C' else 298.15  # 25°C is default T for RH, aw

        psat = vapor_pressure(
            T=T,
            unit=unit,
            source=source,
        )

        if hin == 'p':
            return np.array(val) / (psat * hratio[hout])
        else:  # p is not the input but the output
            return np.array(val) * (psat * hratio[hin])

    else:  # just a conversion between aw and rh
        return np.array(val) * hratio[hin] / hratio[hout]



"""Gathers the formulas for the density of LiCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns two parameters:
- rho0, density of pure water in kg / m^3
- rho, density of solution in kg / m^3
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Conde, M. R., Properties of aqueous solutions of lithium and calcium
  chlorides: formulations for use in air conditioning equipment design.
  International Journal of Thermal Sciences 43, 367-382 (2004).

- Krumgalz, B. S., Pogorelsky, R. & Pitzer, K. S.
  Volumetric Properties of Single Aqueous Electrolytes from Zero to Saturation
  Concentration at 298.15 °K Represented by Pitzer's Ion-Interaction Equations.
  Journal of Physical and Chemical Reference Data 25, 663-689 (1996).
"""

from ...general import SolutionFormula
from ...water.density_atm import DensityAtm_IAPWS

from .misc import relative_rho_conde
from .krumgalz import Density_LiCl_Krumgalz_Base

class Density_LiCl_Conde(SolutionFormula):

    source ='Conde'
    solute = 'LiCl'

    temperature_unit = 'K'
    temperature_range = (273.15, 373.15)

    concentration_unit = 'r'
    concentration_range = (0, 1.273)

    default = True
    with_water_reference = True

    coeffs = 1, 0.540966, -0.303792, 0.100791

    def calculate(self, z, T):
        d = relative_rho_conde(z, self.coeffs)
        density_atm = DensityAtm_IAPWS()
        rho0 = density_atm.calculate(T=T)
        return rho0, rho0 * d


class Density_LiCl_Krumgalz(Density_LiCl_Krumgalz_Base):
    """Already defined in Krumgalz module and not default here"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

Density_LiCl_Formulas = (
    Density_LiCl_Conde,
    Density_LiCl_Krumgalz
)


"""Classes to describe properties containing one or multiple  formulas."""


from .format import format_temperature, format_concentration
from .format import format_output_type

# NOTE: convert() cannot be imported here due to circular import problems


class Property:
    """Base class for properties of water or solutions (e.g. density)"""

    # Below, to define in subclasses
    Formulas = ()  # iterable of formulae available to calculate the property
    quantity = None
    unit = None

    def __repr__(self):
        return f'{self.quantity.capitalize()} {self.unit} (default: {self.default_source})'


class WaterProperty(Property):
    """Generic class for a property that can have various sources"""

    def __init__(self):

        self.formulas = {}  # dict source_name: formula object

        for Formula in self.Formulas:

            formula = Formula()
            source = formula.source
            self.formulas[source] = formula

            if formula.default:
                self.default_source = source

        self.sources = tuple(self.formulas)  # only the source names

    def __call__(self, T=25, unit='C', source=None):
        """Calculate water property as a function of temperature

        Parameters
        ----------
        - T (int, float, array, list, tuple): temperature
        - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin
        - source (str, default None) : Source for the used equation, if None then
        the default source for the particular property is used.
        (see self.sources and self.default_source for more info)

        Output
        ------
        Value of property in SI units
        (float or array of floats depending on input)

        Attributes
        ----------
        .sources --- list of available sources
        .default_source --- source used if None provided
        .quantity --- type of physical quantity (e.g. 'surface tension')
        .unit --- unit of physical quantity (e.g. '[N/m]')
        """
        formula = self.get_formula(source=source)

        T = format_temperature(
            T,
            unit_in=unit,
            unit_out=formula.temperature_unit,
        )

        formula.check_validity_range('temperature', value=T)
        result = formula.calculate(T)
        return format_output_type(result)

    def get_source(self, source=None):
        """Return source if it's in sources, default_source if None."""
        if source is None:
            return self.default_source
        if source in self.sources:
            return source
        raise ValueError(f'Source can only be one of {self.sources}')

    def get_formula(self, source=None):
        """Return formula corresponding to source."""
        source = self.get_source(source=source)
        return self.formulas[source]


class SolutionProperty_Base(Property):
    """Generic class for a property of solutions (saturated or not)"""

    # Define in subclasses (is used to avoid importing convert() in this
    # module, because SolutionProperty is used to defined both the general
    # density and the reduced density used in convert())
    # CAUTION: when putting the converter function as attribute, put the
    # function in staticmethod(), because if not, it will be interpreted as a
    # bound method (and self will be passed as first argument)
    # See e.g. here: https://stackoverflow.com/questions/35321744/python-function-as-class-attribute-becomes-a-bound-method
    converter = None

    def _set_formulas(self, component_type='solute'):
        """Component type is 'solute' or 'crystal' (for saturated solutions)"""

        components = self._get_available_components(component_type)
        setattr(self, f'{component_type}s', components)

        self.formulas = {component: {} for component in components}
        self.sources = {}
        self.default_sources = {}

        for Formula in self.Formulas:

            formula = Formula()
            source = formula.source
            component = getattr(formula, component_type)

            self.formulas[component][source] = formula

            if formula.default:
                self.default_sources[component] = source

            # only the source names
            self.sources[component] = tuple(self.formulas[component])

    def _get_available_components(self, component_type='solute'):
        """Component type is 'solute', or 'crystal' (for saturated solutions)"""
        components = set()
        for Formula in self.Formulas:
            component = getattr(Formula, component_type)
            components.add(component)
        return tuple(components)


class SolutionProperty(SolutionProperty_Base):
    """Generic class for a property of solutions as a function of T/c"""

    # Change in subclasses if needed (e.g. if NaCl not available)
    default_solute = 'NaCl'

    def __init__(self):
        self._set_formulas(component_type='solute')

    def __repr__(self):
        return f'{self.quantity.capitalize()} {self.unit} (solutes: {self.solutes})'

    def get_solute(self, solute=None):
        return self.default_solute if solute is None else solute

    def get_source(self, solute=None, source=None):
        """Return source if it's in sources, default_source if None."""
        solute =self.get_solute(solute=solute)
        if source is None:
            return self.default_sources[solute]
        if source in self.sources[solute]:
            return source
        msg = f'Source can only be one of {self.sources} for {solute}'
        raise ValueError(msg)

    def get_formula(self, solute=None, source=None):
        """Return formula corresponding to source and solute (default if None)"""
        solute = self.get_solute(solute=solute)
        source = self.get_source(source=source, solute=solute)
        return self.formulas[solute][source]

    def __call__(
        self,
        solute=None,
        T=25,
        unit='C',
        relative=False,
        source=None,
        **concentration,
    ):
        """Calculate solution property as a function of temperature and composition

        Parameters
        ----------
        - solute (str): solute name (if None, use default solute)
        - T (float): temperature (default 25)
        - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin
        - relative (bool, default False): True to normalize with pure water at T.
          [ONLY available for some of the properties, e.g. surface tension]

        - source (str, default None) : Source for the used equation, if None then
        gets the default source for the particular solute (defined in submodules).
        See self.sources and self.default_sources for more info

        - **concentration: kwargs with any unit that is allowed by convert(), e.g.
            - m= : molality (mol/kg)
            - w= : mass fraction
            - x= : mole fraction
            - c= : molarity (mol/m^3)
            - r= : mass ratio (unitless)

        Output
        ------
        Value of property in SI units
        (float or array of floats depending on input)

        Attributes
        ----------
        .solutes --- list of available solutes
        .sources --- dict of available sources for every solute
        .default_sources --- default source for every solute
        .quantity --- type of physical quantity (e.g. 'surface tension')
        .unit --- unit of physical quantity (e.g. '[N/m]')
        """
        solute = self.get_solute(solute=solute)
        formula = self.get_formula(source=source, solute=solute)

        if relative and not formula.with_water_reference:
            msg = f'relative=True not available for {formula.quantity} with {solute}'
            raise ValueError(msg)

        T = format_temperature(
            T,
            unit_in=unit,
            unit_out=formula.temperature_unit,
        )

        c = format_concentration(
            concentration=concentration,
            unit_out=formula.concentration_unit,
            solute=solute,
            converter=self.converter,
        )

        formula.check_validity_range('concentration', value=c)
        formula.check_validity_range('temperature', value=T)

        result = formula.calculate(c, T)

        if formula.with_water_reference:
            val_0, val = result
            if relative:
                return format_output_type(val / val_0)
            else:
                return format_output_type(val)

        return format_output_type(result)


class SolutionSolubilityProperty(SolutionProperty_Base):
    """Class for describing solubility"""

    # Change in subclasses if needed (e.g. if NaCl not available)
    default_crystal = 'NaCl'

    def __init__(self):
        self._set_formulas(component_type='crystal')

    def __repr__(self):
        return f'{self.quantity.capitalize()} (crystals: {self.crystals})'

    def get_crystal(self, crystal=None):
        return self.default_crystal if crystal is None else crystal

    def get_source(self, crystal=None, source=None):
        """Return source if it's in sources, default_source if None."""
        crystal =self.get_crystal(crystal=crystal)
        if source is None:
            return self.default_sources[crystal]
        if source in self.sources[crystal]:
            return source
        msg = f'Source can only be one of {self.sources} for {crystal}'
        raise ValueError(msg)

    def get_formula(self, crystal=None, source=None):
        """Return formula corresponding to source and crystal (default if None)"""
        crystal = self.get_crystal(crystal=crystal)
        source = self.get_source(source=source, crystal=crystal)
        return self.formulas[crystal][source]

    def __call__(
        self,
        crystal=None,
        T=25,
        unit='C',
        source=None,
        out='m',
    ):
        """Calculate solution property as a function of temperature and composition

        Parameters
        ----------
        - crystal (str): crystal name (usually same as solute except if several
                                       crystalline forms)
        - T (float): temperature (default 25)
        - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin
        - source (str, default None) : Source for the used equation, if None then
        gets the default source for the particular solute (defined in submodules).
        See self.sources and self.default_sources for more info
        - out: any unit that is allowed by convert(), e.g.
            - m= : molality (mol/kg)
            - w= : mass fraction
            - x= : mole fraction
            - c= : molarity (mol/m^3)
            - r= : mass ratio (unitless)

        Output
        ------
        Value of property in SI units
        (float or array of floats depending on input)

        Attributes
        ----------
        .solutes --- list of available solutes
        .sources --- dict of available sources for every solute
        .default_sources --- default source for every solute
        .quantity --- type of physical quantity (e.g. 'surface tension')
        """
        crystal = self.get_crystal(crystal=crystal)
        formula = self.get_formula(source=source, crystal=crystal)

        T = format_temperature(
            T,
            unit_in=unit,
            unit_out=formula.temperature_unit,
        )

        formula.check_validity_range('temperature', value=T)

        c_unit = formula.concentration_unit
        c_raw = formula.calculate(T)

        c = format_concentration(
            concentration={c_unit: c_raw},
            unit_out=out,
            solute=formula.solute,
            converter=self.converter,
        )

        return format_output_type(c)



"""Gathers the formulas for the activity of KCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns one parameter
- a, water activity (dimensionless, range 0-1)
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Tang, I. N., Munkelwitz, H. R. & Wang, N.
  Water activity measurements with single suspended droplets:
  The NaCl-H2O and KCl-H2O systems.
  Journal of Colloid and Interface Science 114, 409-415 (1986).
  Valid at 25°C and for solutions of molality up to ~13 mol/kg

- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

NOTE: I could not find explicit info about validity domain for the KCl
      formulas in Steiger, so I kept ~ same values as for NaCl
"""

from .steiger import WaterActivity_KCl_Steiger2008_Base
from .tang import WaterActivity_KCl_Tang_Base


class WaterActivity_KCl_Steiger2008(WaterActivity_KCl_Steiger2008_Base):
    """Already defined in steiger module"""
    default = True


class WaterActivity_KCl_Tang(WaterActivity_KCl_Tang_Base):
    """Already defined in tang module"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

WaterActivityFormulas_KCl = (
    WaterActivity_KCl_Steiger2008,
    WaterActivity_KCl_Tang,
)



"""Extension of the properties module for useful shortcut functions

e.g.:
- Kelvin pressure instead of vapor pressure
- molar volume instead of density
etc.
"""

import numpy as np

from . import density_atm, density_sat, surface_tension
from ..format import format_output_type, format_temperature, format_input_type
from ..humidity import format_humidity
from ..constants import Mw, R


# ================================== Config ==================================

rpparams = ['r', 'P']

msg_rp_error = "Input argument can only be 'r=' (radius in m) or 'P='" \
               "(liquid pressure in Pa)"

# ============================================================================


density = {
    'sat': density_sat,
    'atm': density_atm,
}


def molar_volume(T=25, unit='C', source=None, condition='sat'):
    """Molar volume of water [m^3 / mol].

    Parameters
    ----------
    - T (int, float, array, list, tuple): temperature
    - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin
    - source (str, default None) : Source for the used equation, if None then
    the default source for the particular property is used.
    - condition: 'sat' (default): sat. pressure, 'atm': at atm. pressure

    Output
    ------
    Molar volume in m^3 / mol

    Sources
    -------
    see density_atm() and density_sat()
    """
    return Mw / density[condition](T=T, unit=unit, source=source)


# ============== Humidity to kelvin radius and inverse function ==============


def kelvin_pressure(T=25, unit='C', **humidity):
    """Calculate Kelvin (liquid) pressure at a given humidity.

    Parameters
    ----------
    - T: temperature (default 25)
    - unit: temperature unit ('C' or 'K', default 'C')
    - humidity: kwargs p=, rh= or aw=

    Output
    ------
    Kelvin pressure in Pa.

    Examples
    --------
    >>> kelvin_pressure(aw=0.8)  # Kelvin pressure at 80%RH and T=25°C
    -30613102.83763792
    >>> kelvin_pressure(rh=80)           # same
    -30613102.83763792
    >>> kelvin_pressure(p=1000, T=20)    # at 1000Pa, 20°C
    -114763155.07026532
    >>> kelvin_pressure(p=1000, T=293.15, unit='K')    # same
    -114763155.07026532
    >>> kelvin_pressure(aw=[0.5, 0.7, 0.9])  # possible to use iterables
    array([-95092982.94809894, -48932297.94944938, -14454427.57302842])
    """
    aw = format_humidity(unit=unit, T=T, source=None, out='aw', **humidity)
    vm = molar_volume(T=T, unit=unit, source=None)
    Tk = format_temperature(T=T, unit_in=unit, unit_out='K')
    pc = R * Tk * np.log(aw) / vm  # no need to use format_input_type due to np.log
    return format_output_type(pc)


def kelvin_radius(T=25, unit='C', ncurv=2, **humidity):
    """Calculate Kelvin radius at a given humidity.

    Parameters
    ----------
    - T: temperature (default 25)
    - unit: temperature unit ('C' or 'K', default 'C')
    - ncurv: curvature number: 1 cylindrical interface, 2 spherical (default)
    - humidity: kwargs p=, rh= or aw=

    Output
    ------
    Kelvin radius in meters.

    Examples
    --------
    >>> kelvin_radius(aw=0.8)  # Kelvin radius at 80%RH and T=25°C
    4.702052295185309e-09
    >>> kelvin_radius(rh=80)           # same
    4.702052295185309e-09
    >>> kelvin_radius(rh=80, ncurv=1)  # assume cylindrical meniscus instead of spherical
    2.3510261475926545e-09
    >>> kelvin_radius(p=1000, T=20)    # at 1000Pa, 20°C
    1.2675869773199224e-09
    >>> kelvin_radius(p=1000, T=293.15, unit='K')    # same
    1.2675869773199224e-09
    >>> kelvin_radius(aw=[0.5, 0.7, 0.9])  # possible to use iterables
    array([1.51372274e-09, 2.94170551e-09, 9.95849955e-09])
    """
    sig = surface_tension(T=T, unit=unit)
    pc = kelvin_pressure(T=T, unit=unit, **humidity)
    r = -ncurv * sig / pc
    return format_output_type(r)


def kelvin_humidity(T=25, unit='C', ncurv=2, out='aw', **r_or_p):
    """Calculate humidity corresponding to a Kelvin radius.

    Parameters
    ----------
    - r: Kelvin radius in meters
    - T: temperature (default 25)
    - unit: temperature unit ('C' or 'K', default 'C')
    - ncurv: curvature number: 1 cylindrical interface, 2 spherical (default)
    - out: type of output ('p', 'rh', or 'aw')
    - input: kwargs (r= for radius, or p= for liquid pressure)

    Output
    ------
    Kelvin radius in meters.

    Examples
    --------
    # With radius in meters as input -----------------------------------------
    >>> kelvin_humidity(r=4.7e-9)  # activity corresponding to Kelvin radius of 4.7 nm at 25°C
    0.7999220537658477
    >>> kelvin_humidity(r=4.7e-9, out='rh')  # same, but expressed in %RH instead of activity
    79.99220537658476
    >>> kelvin_humidity(r=4.7e-9, out='p')  # same, but in terms of pressure (Pa)
    2535.612513169546
    >>> kelvin_humidity(r=4.7e-9, out='p', T=293.15, unit='K')  # at a different temperature
    1860.0699544036922
    >>> kelvin_humidity(r=4.7e-9, ncurv=1)  # cylindrical interface
    0.8943836166689592
    >>> kelvin_humidity(r=[3e-9, 5e-9])  # with iterables
    array([0.70486836, 0.81070866])

    # Similar examples, but with liquid pressure (Pa) as input ---------------
    >>> kelvin_humidity(P=-30e6)  # humidity corresponding to P=-30 MPa
    0.8035832003814989
    >>> kelvin_humidity(P=-30e6, out='rh')
    80.35832003814988
    >>> kelvin_humidity(P=-30e6, out='p', T=293.15, unit='K')
    1873.2224722706874
    >>> kelvin_humidity(P=[-30e6, -50e6])
    array([0.8035832 , 0.69457329])
    """
    try:
        rpin, = r_or_p.keys()    # radius or pressure keyword
        val, = r_or_p.values()  # check there is only one input humidity arg.
    except ValueError:
        raise KeyError(msg_rp_error)

    if rpin not in rpparams:
        raise KeyError(msg_rp_error)

    if rpin == 'r':
        r = format_input_type(val)
        sig = surface_tension(T=T, unit=unit)
        pc = - ncurv * sig / r
    else:
        pc = format_input_type(val)

    vm = molar_volume(T=T, unit=unit)
    Tk = format_temperature(T=T, unit_in=unit, unit_out='K')
    aw = np.exp(pc * vm / (R * Tk))
    hout = format_humidity(unit=unit, T=T, source=None, out=out, aw=aw)
    return format_output_type(hout)



"""Parameter conversions for water and solutions. Only basic ones (w, x, m).

Basic parameters do not rely on volumetric and temperature dependent quantities
such as density.

One reason for this is to prevent circular imports
(e.g. more elaborate units such as molarity require knowledge about density,
and density uses basic conversions in its core)
Also, fancier things such as ionic strength are easier to treat separately

SOURCES
-------
CRC Handbook - Conversion Formulas for Concentration of Solutions
http://hbcponline.com/faces/documents/01_19/01_19_0001.xhtml
"""

# TODO -- write more comprehensive examples
# TODO -- add checks that inputs are within normal range (e.g. 0-1 for x or w)

from ...constants import molar_mass, Mw, solute_list
from ...format import check_solute, check_units

# Define what solutes and units are acceptable for conversion


allowed_units = ['x', 'w', 'm', 'r']


# ========================= MAIN CONVERSION FUNCTION =========================

def basic_convert(value, unit1, unit2, solute='NaCl'):
    """Convert between concentrations, molalities etc. for solutions.

    Parameters
    ----------
    - value (float): value to convert
    - unit1 (str): its unit.
    - unit2 (str): unit to convert to.
    - solute (str): name of solute, (default 'NaCl').

    solute has to be in the solute list in the constants module
    unit1 and unit2 have to be in the allowed units list :
    'x' (mole fraction.), 'w' (weight frac.), 'm' (molality), 'r' (mass ratio)

    Output
    ------
    Converted value (dimensionless or SI units)

    Examples
    --------
    - convert(0.4, 'w', 'x'): mass fraction of 0.4 into mole fraction for NaCl
    - convert(10, 'm', 'w'): molality of 10 mol/kg into mass fraction for NaCl
    - convert(10, 'm', 'w', 'LiCl'): same but for LiCl.
    """

    # No need to calculate anything if the in and out units are the same -----
    if unit1 == unit2:
        return value

    check_solute(solute, solute_list)
    check_units([unit1, unit2], allowed_units)

    # Get useful quantities related to solute --------------------------------
    M = molar_mass(solute)         # molar mass of solute, kg/mol

    if unit1 == 'w':  # weight fraction to other quantities ------------------

        w = value

        if unit2 == 'x':
            return (w / M) / (w / M + (1 - w) / Mw)

        elif unit2 == 'm':
            return w / ((1 - w) * M)

        elif unit2 == 'r':
            return w / (1 - w)

    if unit1 == 'x':  # mole fraction to other quantities --------------------

        x = value

        if unit2 == 'w':
            return x * M / (x * M + (1 - x) * Mw)

        elif unit2 == 'm':
            return 1 / Mw * x / (1 - x)

        elif unit2 == 'r':
            return M / Mw * x / (1 - x)

    if unit1 == 'm':   # molality to other quantities ------------------------

        m = value

        if unit2 == 'w':

            return m * M / (1 + m * M)

        elif unit2 == 'x':
            return m * Mw / (1 + Mw * m)

        elif unit2 == 'r':
            return M * m

    if unit1 == 'r':  # mass ratio to other quantities --------------

        z = value

        if unit2 == 'w':
            return z / (1 + z)

        elif unit2 == 'x':
            return z / M / (z / M + 1 / Mw)

        elif unit2 == 'm':
            return z / M



"""Gathers the formulas for the solubility of LiCl solutions.

Sources
-------
- CRC Handbook of Chemistry and Physics, 104th Edition
  Section: 4 | Solubility of Common Inorganic Salts as a Function of Temperature
  https://hbcp.chemnetbase.com/documents/04_29/04_29_0001.xhtml?dswid=7662
"""

from .crc_handbook import Solubility_LiCl_CRCHandbook_Base


class Solubility_LiCl_CRCHandbook(Solubility_LiCl_CRCHandbook_Base):
    """Already defined in CRC Handbook module"""
    default = True


# ============================= WRAP-UP FORMULAS =============================

SolubilityFormulas_LiCl = (
    Solubility_LiCl_CRCHandbook,
)


"""Activity coefficients of solutions"""

from .KCl import SolubilityFormulas_KCl
from .LiBr import SolubilityFormulas_LiBr
from .LiCl import SolubilityFormulas_LiCl
from .Na2SO4 import SolubilityFormulas_Na2SO4
from .NaCl import SolubilityFormulas_NaCl

SolubilityFormulas = (
    SolubilityFormulas_KCl +
    SolubilityFormulas_LiBr +
    SolubilityFormulas_LiCl +
    SolubilityFormulas_Na2SO4 +
    SolubilityFormulas_NaCl
)



"""Various tools to calculate solution activities using Steiger formulae."""

import json
from copy import deepcopy
from pathlib import Path

import numpy as np

from ...constants import dissociation_numbers


# corrective factors between 2005 / 2008 for 1:1 and 2:1 electrolytes
c_11 = 2
c_21 = 4 / 2**(1 / 2)

# Correspondence between number of ions and corrective factor
pitzer_corr = {2: c_11, 3: c_21}


class CoeffsSteiger2005_Activity:
    """Load and manage Steiger 2005 coefficients for activity"""

    file = 'Coeffs_Steiger2005_Activity.json'

    def __init__(self):
        """Load and format raw coefficients stored in json file."""
        path = Path(__file__).parent / 'data' / self.file
        with open(path, 'r', encoding='utf8') as file:
            self.all_coeffs = json.load(file)

    def coeffs(self, solute='NaCl', T=298.15):
        """Get beta, C, A_phi, etc. coefficients at 298.15K for a given solute"""
        coeffs = self.all_coeffs[solute]
        coeffs['A_phi'] = self.all_coeffs['A_phi']
        return coeffs


class CoeffsSteiger2008:
    """Load and manage Steiger 2008 coefficients for activity and solubility

    (base class)
    """

    file = None  # define in subclasses

    def __init__(self):
        """Load and format raw coefficients stored in json file."""
        path = Path(__file__).parent / 'data' / self.file
        with open(path, 'r', encoding='utf8') as file:
            raw_coeffs = json.load(file)
        self.all_coeffs = self._format_coeffs(raw_coeffs)

    @staticmethod
    def _calculate_parameter(T, *coeffs):
        """Determine betas or third virial coefficient C as a function of T.

        Note: is also used to calculate solubility as a function of T.
        (see solutions.saturated)
        """
        Tr = 298.15
        q1, q2, q3, q4, q5, q6 = coeffs
        term2 = q2 * (1 / T - 1 / Tr)
        term3 = q3 * np.log(T / Tr)
        term4 = q4 * (T - Tr)
        term5 = q5 * (T ** 2 - Tr ** 2)
        term6 = q6 * np.log(T - 225)
        return q1 + term2 + term3 + term4 + term5 + term6


class CoeffsSteiger2008_Activity(CoeffsSteiger2008):
    """Load and manage Steiger 2008 coefficients for activity"""

    file = 'Coeffs_Steiger2008_Activity.json'

    @staticmethod
    def _format_coeffs(raw_coeffs):
        """Take into account corrective factors depending on dissociation numbers."""
        all_coeffs = deepcopy(raw_coeffs)
        A_phi = all_coeffs.pop('A_phi')
        for solute, coeffs in all_coeffs.items():
            c_phi = coeffs.pop('C_phi')
            nu_mx = sum(dissociation_numbers[solute])
            corr = pitzer_corr[nu_mx]
            coeffs['C_phi'] = [corr * q for q in c_phi]
        all_coeffs['A_phi'] = A_phi
        return all_coeffs

    def A_phi(self, T=298.15):
        """Debye-Huckel parameter as a function of the temperature T at 0.1MPa"""
        a1, a2, a3, a4, a5, a6 = self.all_coeffs['A_phi']
        return a1 + a2 / (T - 222) + a3 / T**2 + a4 * T + a5 * T**2 + a6 * T**4

    def coeffs(self, solute='NaCl', T=298.15):
        """Get beta, C, A_phi, etc. coefficients ata a given temperature for a given solute"""
        coeffs = {}
        coeffs['A_phi'] = self.A_phi(T)
        for name in 'beta0', 'beta1', 'beta2', 'C_phi':
            qs = self.all_coeffs[solute][name]
            coeffs[name] = self._calculate_parameter(T, *qs)
        for name in 'alpha1', 'alpha2':
            coeffs[name] = self.all_coeffs[solute][name]
        return coeffs


class CoeffsSteiger2008_Solubility(CoeffsSteiger2008):
    """Load and manage Steiger 2008 coefficients for solubility"""

    file = 'Coeffs_Steiger2008_Solubility.json'

    @staticmethod
    def _format_coeffs(raw_coeffs):
        """Take into account corrective factors depending on dissociation numbers."""
        return raw_coeffs

    def ln_K(self, crystal='NaCl', T=298.15):
        return self._calculate_parameter(T, *self.all_coeffs[crystal])


coeffs_steiger2005_activity = CoeffsSteiger2005_Activity()
coeffs_steiger2008_activity = CoeffsSteiger2008_Activity()
coeffs_steiger2008_solubility = CoeffsSteiger2008_Solubility()



"""Extension of formulas from the properties of aqueous solutions.

e.g., osmotic_pressure instead of water_activity, etc.
"""

import numpy as np

from ..constants import R, Mw, dissociation_numbers
from ..format import format_temperature, format_output_type, format_concentration
from ..water import molar_volume

from .convert import convert
from .properties import water_activity, solubility


def osmotic_pressure(solute='NaCl', T=25, unit='C', source=None, **concentration):
    """Return osmotic pressure of an aqueous solution at a given concentration.

    Basically the same as water_activity, but expressed as -RT/vm * log(a_w)

    Parameters
    ----------
    - solute (str): solute name, default 'NaCl'
    - T (float): temperature (default 25)
    - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin

    - source (str, default None) : Source for the used equation, if None then
    gets the default source for the particular solute (defined in submodules).
    See summary of available sources below.

    - **concentration: kwargs with any unit that is allowed by convert(), e.g.
        - m= : molality (mol/kg)
        - w= : mass fraction
        - x= : mole fraction
        - c= : molarity (mol/m^3)
        - r= : mass ratio (unitless)

    Output
    ------
    - Osmotic pressure (Π) in Pa

    Solutes and Sources
    -------------------
    See water_activity.solutes, water_activity.sources and
    water_activity.default_sources and README.md

    Examples
    --------
    - osmotic_pressure(x=0.1) returns Π for a mole fraction of 0.1 of NaCl
    - osmotic_pressure(w=0.2) returns Π for a mass fraction of 0.2 of NaCl
    - osmotic_pressure(c=5000) returns Π for a molality of 5 mol/L of NaCl
    - osmotic_pressure(m=6) returns Π for a molality of 6 mol/kg of NaCl
    - osmotic_pressure('LiCl', m=6): same for LiCl
    - osmotic_pressure('LiCl', m=6, T=30): same for LiCl at 30°C
    - osmotic_pressure('LiCl', 293, 'K', m=6): same for LiCl at 293K.
    - osmotic_pressure(solute='CaCl2', T=50, m=[2, 4, 6])  # concentration as iterable
    """
    a_w = water_activity(
        solute=solute,
        T=T,
        unit=unit,
        source=source,
        **concentration,
    )
    T_kelvin = format_temperature(T=T, unit_in=unit, unit_out='K')
    pi = - R * T_kelvin / molar_volume(T=T_kelvin, unit='K') * np.log(a_w)
    return format_output_type(pi)


def osmotic_coefficient(solute='NaCl', T=25, unit='C', source=None, **concentration):
    """Return osmotic coefficient (Φ) of an aqueous solution at a given concentration.

    Basically the same as water_activity, but expressed as Φ

    Parameters
    ----------
    - solute (str): solute name, default 'NaCl'
    - T (float): temperature (default 25)
    - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin

    - source (str, default None) : Source for the used equation, if None then
    gets the default source for the particular solute (defined in submodules).
    See summary of available sources below.

    - **concentration: kwargs with any unit that is allowed by convert(), e.g.
        - m= : molality (mol/kg)
        - w= : mass fraction
        - x= : mole fraction
        - c= : molarity (mol/m^3)
        - r= : mass ratio (unitless)

    Output
    ------
    - Osmotic coefficient (Φ), dimensionless

    Solutes and Sources
    -------------------
    See water_activity.solutes, water_activity.sources and
    water_activity.default_sources and README.md

    Examples
    --------
    - osmotic_coefficient(x=0.1) returns Φ for a mole fraction of 0.1 of NaCl
    - osmotic_coefficient(w=0.2) returns Φ for a mass fraction of 0.2 of NaCl
    - osmotic_coefficient(c=5000) returns Φ for a molality of 5 mol/L of NaCl
    - osmotic_coefficient(m=6) returns Φ for a molality of 6 mol/kg of NaCl
    - osmotic_coefficient('LiCl', m=6): same for LiCl
    - osmotic_coefficient('LiCl', m=6, T=30): same for LiCl at 30°C
    - osmotic_coefficient('LiCl', 293, 'K', m=6): same for LiCl at 293K.
    - osmotic_coefficient(solute='CaCl2', T=50, m=[2, 4, 6])  # concentration as iterable
    """
    a_w = water_activity(
        solute=solute,
        T=T,
        unit=unit,
        source=source,
        **concentration,
    )
    m = format_concentration(
        concentration=concentration,
        unit_out='m',
        solute=solute,
        converter=convert,
    )
    nu_mx = sum(dissociation_numbers[solute])
    phi = - np.log(a_w) / (Mw * nu_mx * m)
    return format_output_type(phi)


def aw_saturated(
    crystal='NaCl',
    T=25,
    unit='C',
    activity_source=None,
    solubility_source=None,
):
    """Water activity of the saturated solution.

    Basically the same as solubility, but expressed as a_w

    Parameters
    ----------
    - crystal (str): crystal name, default 'NaCl'
    - T (float): temperature (default 25)
    - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin

    - source (str, default None) : Source for the used equation, if None then
    gets the default source for the particular solute (defined in submodules).
    See summary of available sources below.

    Output
    ------
    - Water activity, dimensionless

    Solutes and Sources
    -------------------
    See solubility.solutes, solubility.sources and
    solubility.default_sources and README.md

    Examples
    --------
    - aw_saturated()               # of NaCl at 25°C
    - aw_saturated('LiCl', T=15)   # of LiCl at 15°C
    """
    m_sat = solubility(
        crystal=crystal,
        T=T,
        unit=unit,
        source=solubility_source,
        out='m',
    )

    formula = solubility.get_formula(
        crystal=crystal,
        source=solubility_source,
    )

    a_w = water_activity(
        solute=formula.solute,
        T=T,
        unit=unit,
        source=activity_source,
        m=m_sat,
    )
    return a_w


"""Refractive index of solutions according to Tan

Source
------
- Tan (default) : "Dependence of Refractive Index on Concentration and
Temperature in Electrolyte Solution, Polar Solution, Nonpolar Solution, and
Protein Solution", Tan & Huang, J. Chem. Eng. Data  (2015).
Valid from w = 0 to w = 0.25 and for temperatures between 20 and 45°C
"""

from ...general import SolutionFormula


class RefractiveIndex_Tan_Base(SolutionFormula):

    source ='Tan'

    temperature_unit = 'C'
    temperature_range = (20, 45)

    concentration_unit = 'w'

    with_water_reference = False

    def calculate(self, w, T):
        c = w * 100   # avoid using *= to not mutate objects in place
        n0, a1, a2, b1, b2 = self.coeffs.values()
        return n0 + a1 * c + a2 * c**2 + b1 * T + b2 * T**2


class RefractiveIndex_CaCl2_Tan_Base(RefractiveIndex_Tan_Base):
    solute = 'CaCl2'
    concentration_range = (0, 0.15)
    coeffs = {
        'n0': 1.3373,
        'a1': 2.5067e-3,
        'a2': -3.9e-8,
        'b1': -1.1122e-4,
        'b2': -4e-9,
    }


class RefractiveIndex_KCl_Tan_Base(RefractiveIndex_Tan_Base):
    solute = 'KCl'
    concentration_range = (0, 0.15)
    coeffs = {
        'n0': 1.3352,
        'a1': 1.6167e-3,
        'a2': -4e-7,
        'b1': -1.1356e-4,
        'b2': -5.7e-9,
    }


class RefractiveIndex_NaCl_Tan_Base(RefractiveIndex_Tan_Base):
    solute = 'NaCl'
    concentration_range = (0, 0.25)
    coeffs = {
        'n0': 1.3373,
        'a1': 1.7682e-3,
        'a2': 0,          # modified -5.8e-6 to 0 to fit the data adequately
        'b1': -1.3531e-4,
        'b2': -5.1e-8,
    }



"""Classes to describe specific formulas for water and solutions"""

from warnings import warn

import numpy as np


class Formula:
    """base class for formulas for water/solution properties"""

    # To be defined in subclasses
    source =''
    input_types = ()
    default = False  # Change to True to select default source

    def __repr__(self):
        msg = f'{self.source} formula'
        if self.default:
            msg += ' [default]'
        msg += '; '
        for input_type in self.input_types:
            validity_info = self._get_range_and_unit(input_type)
            v1, v2 = validity_info['range']
            unit = validity_info['unit']
            if v1 == v2:
                value = f'{v1}[{unit}]'
            else:
                value = f'{v1}-{v2}[{unit}]'
            msg += f"{input_type} {value} "
        return msg

    def _get_range_and_unit(self, input_type):
        if input_type not in self.input_types:
            raise ValueError(f'Input type {input_type} not in {self.input_types}')
        okrange = getattr(self, f'{input_type}_range')
        unit = getattr(self, f'{input_type}_unit')
        return {'range': okrange, 'unit': unit}

    def check_validity_range(self, input_type, value):
        """Check value is in validity range, issues warning (no error) if not.

        Parameters
        ----------
        - value_type: 'temperature' or 'concentration'
        - value (scalar, list, array, tuple etc.), in the same unit as okrange.

        Optional (used only for the warning to be explicit):
        - dataname (str): name of the parameter (e.g. 'temperature'),
        - unitname (str): unit of the source data (e.g. '°C' or 'x')
        - sourcename (str): name of the source of the data
        """
        validity_info = self._get_range_and_unit(input_type=input_type)
        val_min, val_max = validity_info['range']
        unit = validity_info['unit']

        try:  # This works only if value is a single value, not an array or a list
            out_of_range = value < val_min or value > val_max
        except ValueError:  # if array, list, array, tuple etc, transform to 1D np array
            values = np.array(value).flatten()
            out_of_range = any(values < val_min) or any(values > val_max)

        if out_of_range:
            warn(
                f'{input_type.capitalize()} outside of validity range'
                f'({unit} in [{val_min}-{val_max}]) for {self.source}.'
                f'[{self.solute}]',
                stacklevel=2
            )

    def calculate(self, *args, **kwargs):
        """To define in subclasses"""
        pass


class WaterFormula(Formula):
    """Formulas for water properties as a function of T"""

    # To be defined in subclasses
    temperature_unit = None
    temperature_range = None

    # Do not change below
    input_types = 'temperature',


class SolutionFormula(Formula):
    """Formulas for solution properties as a function of T and composition"""

    # To be defined in subclasses --------------------------------------------

    solute = None

    # if True, formula returns val(c=0) and val(c) to allow for relative calculations
    with_water_reference = False

    temperature_unit = None
    temperature_range = None
    concentration_unit = None
    concentration_range = None

    # Do not change below ----------------------------------------------------

    input_types = 'temperature', 'concentration'


class SaturatedSolutionFormula(Formula):
    """Formulas for saturated solution properties (depend only on T)"""

    # To be defined in subclasses --------------------------------------------
    crystal = None   # because several crystal forms are possible
    crystal_hydration = None  # put integer if crystal is hydrated
    solute = None  # if not provided, extracted automatically from cystal name

    temperature_unit = None
    temperature_range = None

    # Do not change below ----------------------------------------------------

    input_types = 'temperature',

    def __init__(self):
        super().__init__()
        if self.solute is None:
            components = self.crystal.split(',')
            if len(components) > 2:
                raise ValueError(f'Crystal {self.crystal} not recognized')
            elif len(components) == 1:
                self.solute = self.crystal
            else:
                solute, _ = components
                self.__class__.solute = solute



"""Activity of solutions according to clegg

Source
------
- Clegg et al. : "Thermodynamic Properties of Aqueous Aerosols to High
Supersaturation: II" (1997). Valid at 25°C and for solutions of molality
up to ~17 mol/kg (x ~ 0.23)
"""

import numpy as np

from ....constants import charge_numbers
from ...general import SolutionFormula
from ..ionic import ion_quantities, ionic_strength


def aw_clegg(x, T, solute, coeffs):

    x_ion1, x_ion2 = ion_quantities(solute, x=x)
    x1 = 1 - (x_ion1 + x_ion2)  # mole fraction of water

    z_ion1, z_ion2 = charge_numbers[solute]

    Ix = ionic_strength(solute, x=x)  # ionic strength

    rho = 13.0

    A_x, B, alpha, W1, U1, V1 = coeffs

    val = 2 * A_x * Ix**(3 / 2) / (1 + rho * Ix**(1 / 2))  # 1st line
    val -= x_ion1 * x_ion2 * B * np.exp(-alpha * Ix**(1 / 2))  # 2nd line
    val += (1 - x1) * x_ion2 * (1 + z_ion2) * W1  # 5th line
    val += (1 - 2 * x1) * x_ion1 * x_ion2 * ((1 + z_ion2)**2 / z_ion2) * U1  # 6-7th lines
    val += 4 * x1 * (2 - 3 * x1) * x_ion1 * x_ion2 * V1  # 8th line

    f1 = np.exp(val)
    a1 = f1 * x1

    return a1


class WaterActivity_Clegg_Base(SolutionFormula):

    source ='Clegg'

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'x'

    with_water_reference = False

    def calculate(self, x, T):
        return aw_clegg(x, T, solute=self.solute, coeffs=self.coeffs.values())


class WaterActivity_Na2SO4_Clegg_Base(WaterActivity_Clegg_Base):

    solute = 'Na2SO4'
    concentration_range = (0, 0.23)

    coeffs = {
        'A_x': 2.915,
        'B': 48.56028,
        'alpha': 8.0,
        'W1': 5.555706,
        'U1': 21.88352,
        'V1': -22.81674,
    }


class WaterActivity_NaCl_Clegg_Base(WaterActivity_Clegg_Base):

    solute = 'NaCl'
    concentration_range = (0, 0.25)

    coeffs = {
        'A_x': 2.915,
        'B': 24.22023,
        'alpha': 5.0,
        'W1': 0.7945378,
        'U1': 12.15304,
        'V1': -12.76357,
    }



"""Surface tension of solutions"""

from .CaCl2 import SurfaceTensionFormulas_CaCl2
from .KCl import SurfaceTensionFormulas_KCl
from .LiCl import SurfaceTensionFormulas_LiCl
from .MgCl2 import SurfaceTensionFormulas_MgCl2
from .Na2SO4 import SurfaceTensionFormulas_Na2SO4
from .NaCl import SurfaceTensionFormulas_NaCl

SurfaceTensionFormulas = (
    SurfaceTensionFormulas_CaCl2 +
    SurfaceTensionFormulas_KCl +
    SurfaceTensionFormulas_LiCl +
    SurfaceTensionFormulas_MgCl2 +
    SurfaceTensionFormulas_Na2SO4 +
    SurfaceTensionFormulas_NaCl
)



"""Vapor pressure of water as a function of temperature.

Sources
-------

--- 'IAPWS':
    W. Wagner, A. Pruß
    "The IAPWS Formulation 1995 for the Thermodynamic Properties of Ordinary
    Water Substance for General and Scientific Use"
    Journal of Physical and Chemical Reference Data
    (2002)
*Note*: Psat expression from IAPWS; temperature validity range seems
        to be 0 - 1000°C. Equation is (2.5) page 398.

--- 'Wexler':
    A. Wexler, L. Greenspan
    "Vapor Pressure Equation for Water in the Range 0 to 100°C"
    Journal of Research of the National Bureau of Standards - A. Physics and
    Chemistry
    (1971)
*Note*: Psat expression, valid from 0 to 100°C.
        We use Equation (17) and not the simplified Equations 18a-c.'

--- 'Bridgeman':
    O. C. Bridgeman, E. W. Aldrich
    "Vapor Pressure Tables for Water"
    Journal of Heat Transfer
    (1964)
*Note*: Psat expression, valid from 0 to 374.15°C.

Notes
-----
Some older references use different scales of temperatures (e.g. IPS68 / 90)
which might be why some values for the critical point are slightly different.
Here we consider all absolute temperatures as being in Kelvin.
"""

import numpy as np

from ...constants import Tc, Patm, Pc, Tc
from ..general import WaterFormula


class VaporPressure_IAPWS(WaterFormula):
    """Water saturation pressure according to Wagner & Pruß, T in K."""

    source ='IAPWS'
    temperature_unit = 'K'
    temperature_range = (273.15, Tc)  # in fact 273.16 (triple point)
    default = True

    coeffs = [
        -7.85951783,
        1.84408259,
        -11.7866497,
        22.6807411,
        -15.9618719,
        1.80122502
    ]

    def calculate(self, T):
        v = 1 - T / Tc
        a1, a2, a3, a4, a5, a6 = self.coeffs

        val = Tc / T * (a1 * v + a2 * v**1.5 + a3 * v**3 + a4 * v**3.5 + a5 * v**4 + a6 * v**7.5)

        return np.exp(val) * Pc


class VaporPressure_Wexler(WaterFormula):
    """Water Saturation pressure according to Wexler 1971, eq. (17). T in K."""

    source ='Wexler'
    temperature_unit = 'K'
    temperature_range = (273.15, 373.15)

    coeffs = {
        'E':
            [
                -7.51152e3,
                9.65389644e1,
                2.399897e-2,
                -1.1654551e-5,
                -1.2810336e-8,
                2.0998405e-11,
            ],
        'B': -1.2150799e1
    }

    def calculate(self, T):
        lnp = self.coeffs['B'] * np.log(T)
        for i, e in enumerate(self.coeffs['E']):
            lnp += e * T**(i - 1)
        return np.exp(lnp)


class VaporPressure_Bridgeman(WaterFormula):
    """Water Saturation pressure according to Bridgeman 1964. T in C."""

    source ='Bridgeman'
    temperature_unit = 'C'
    temperature_range = (0, 374.15)

    coeffs = {
        'A': 1.06423320,
        'B': 1.0137921,
        'C': 5.83531e-4,
        'D': 4.16385282,
        'E': 237.098157,
        'F': 0.30231574,
        'G': 3.377565e-3,
        'H': 1.152894,
        'K': 0.745794,
        'L': 654.2906,
        'M': 266.778,
        'Tx': 187,
    }

    def calculate(self, T):
        A, B, C, D, E, F, G, H, K, L, M, Tx = self.coeffs.values()

        Y1 = D * (T - Tx) / (T + E)
        X = (T - Tx) / 100
        Z = Tx / 100 * (-1 + 2 * (H - K * np.arccosh(L / (T + M))))
        a = Z**2 * ((Tx / 100)**2 - Z**2) / (F * (1 + G * T))

        yfact = 3 * np.sqrt(3) / (2 * (Tx / 100)**3)
        Y2 =  yfact * (X - 0.01 * a) * ((Tx / 100)**2 - (X - 0.01 * a)**2) / 100

        logp = A + Y1 - B * (1 + C * T) * Y2

        return (10**logp) * Patm


# ========================== WRAP-UP OF FORMULAS =============================

VaporPressureFormulas = (
    VaporPressure_IAPWS,
    VaporPressure_Wexler,
    VaporPressure_Bridgeman,
)



"""Gathers the formulas for the solubility of LiBr solutions.

Sources
-------
--- Duvall
Duvall, K. N., Dirksen, J. A. & Ring, T. A.
Ostwald-Meyers Metastable Region in LiBr Crystallization—Comparison of
Measurements with Predictions.
Journal of Colloid and Interface Science 239, 391-398
(2001)
"""

from .duvall import Solubility_LiBr_1H2O_Duvall_Base
from .duvall import Solubility_LiBr_2H2O_Duvall_Base
from .duvall import Solubility_LiBr_3H2O_Duvall_Base


class Solubility_LiBr_1H2O_Duvall(Solubility_LiBr_1H2O_Duvall_Base):
    """Default for monohydrate"""
    default = True


class Solubility_LiBr_2H2O_Duvall(Solubility_LiBr_2H2O_Duvall_Base):
    """Default for monohydrate"""
    default = True


class Solubility_LiBr_3H2O_Duvall(Solubility_LiBr_3H2O_Duvall_Base):
    """Default for monohydrate"""
    default = True


# ============================= WRAP-UP FORMULAS =============================

SolubilityFormulas_LiBr = (
    Solubility_LiBr_1H2O_Duvall,
    Solubility_LiBr_2H2O_Duvall,
    Solubility_LiBr_3H2O_Duvall,
)



"""Plot properties of water and solutions."""

# General imports (non-standard)
import matplotlib.pyplot as plt
import numpy as np

# Local imports
from .format import format_temperature, format_concentration

from .water import vapor_pressure, surface_tension as sigma_w
from .water import density_atm, density_sat
from .water import diffusivity_in_air, viscosity_atm

from .solutions import water_activity
from .solutions import activity_coefficient
from .solutions import surface_tension as sigma_s
from .solutions import density
from .solutions import refractive_index, electrical_conductivity
from .solutions import solubility
from .solutions import convert


npts = 200

temperature_unit = 'C'
concentration_unit = 'm'

LINESTYLES = [
    '-',
    '--',
    '-.',
    ':',
    (5, (10, 3)),             # long dash with offset
    (0, (3, 1, 1, 1, 1, 1)),  # densely dashdotdotted
    (0, (3, 5, 1, 5, 1, 5)),  # dashdotdotted
]


# ================================ WATER =====================================


faxs = plt.subplots(2, 3)
fig_w, ((ax_w_psat, ax_w_sigma, ax_w_rho), (ax_w_diff, ax_w_visc, _)) = faxs

fig_w.suptitle('Water')

water_properties = (
    vapor_pressure,
    sigma_w,
    density_sat,
    density_atm,
    diffusivity_in_air,
    viscosity_atm,
)

# General plotting functions -------------------------------------------------

def plot_all_sources(ppty, ax, norm=1):
    """Plot all available sources for a given property

    Inputs
    ------
    ppty: property object/function (e.g. surface_tension, vapor_pressure etc.)
    ax: Matplotlib axes in which to plot the data
    norm (float): normalization factor for plotting the property (default 1)
    """
    for source, linestyle in zip(ppty.sources, LINESTYLES):

        formula = ppty.get_formula(source)

        tmin, tmax = formula.temperature_range
        unit = formula.temperature_unit

        tt_raw = np.linspace(tmin, tmax, npts)
        tt = format_temperature(tt_raw, unit, temperature_unit)

        data = ppty(T=tt, unit=temperature_unit, source=source)
        ax.plot(tt, data * norm, ls=linestyle, label=source)

    ax.legend()
    ax.set_xlabel(f'T ({temperature_unit})')
    ax.set_ylabel(f'{ppty.quantity.capitalize()} {ppty.unit}')


# Plots
plot_all_sources(vapor_pressure, ax_w_psat, norm=1e-3)
ax_w_psat.set_ylabel('Vapor pressure [kPa]')

plot_all_sources(sigma_w, ax_w_sigma, norm=1e3)
ax_w_sigma.set_ylabel('Surface tension [mN/m]')

plot_all_sources(density_sat, ax_w_rho)
plot_all_sources(density_atm, ax_w_rho)
ax_w_rho.set_ylabel('Density [kg / m^3]')

plot_all_sources(diffusivity_in_air, ax_w_diff)
plot_all_sources(viscosity_atm, ax_w_visc)


# # ============================== SOLUTIONS ===================================


solution_properties = (
    activity_coefficient,
    water_activity,
    density,
    sigma_s,
    refractive_index,
    electrical_conductivity,
)

COLORS = {
    'NaCl': 'cornflowerblue',
    'KCl': 'darkblue',
    'LiBr': 'rebeccapurple',
    'LiCl': 'lightblue',
    'KI': 'darkgreen',
    'MgCl2': 'gold',
    'CaCl2': 'orange',
    'Na2SO4': 'brown',
}


# General plotting functions -------------------------------------------------

def plot_all_sources_conc(
    ppty,
    solute,
    T=25,
    unit='C',
    ctype='m',
    ax=None,
    norm=1,
    linestyle=None,
    color=None,
):
    """Plot all available sources for a given property/solute as a function of concentration

    Inputs
    ------
    ppty: property object/function (e.g. surface_tension, density etc.)
    solute (str): name of solute (e.g. 'NaCl')
    T, unit : temperature and temperature unit
    ctype: unit of concentration to plot the data (e.g. 'x', 'm', 'c' etc.)
    relative (bool): if True, use the relative option when applicable
    ax: Matplotlib axes in which to plot the data
    norm (float): normalization factor for plotting the property (default 1)
    linestyle, color: if not None, override default COLORS/LINESTYLES
    """
    for source, ls in zip(ppty.sources[solute], LINESTYLES):
        formula = ppty.get_formula(solute=solute, source=source)

        cmin, cmax = formula.concentration_range
        cunit = formula.concentration_unit

        # The 0.999... is to avoid rounding making data out of range
        if source == 'Clegg' and ppty.quantity == 'density':
            c_max = cmax * 0.75
        else:
            c_max = cmax * 0.999

        c_min = cmin if cmin == 0 else 1.001 * cmin

        cc_raw = np.linspace(c_min, c_max, npts)

        cc = format_concentration(
            concentration={cunit: cc_raw},
            unit_out=ctype,
            solute=solute,
            converter=convert,
        )

        kwargs = {
            'solute': solute,
            'T': T,
            'unit': unit,
            'source': source,
            ctype: cc,
        }

        data = ppty(**kwargs)

        name = f"{solute}, {source} (T={T})"

        plot_kwargs = {'label': name}
        plot_kwargs['color'] = COLORS[solute] if color is None else color
        plot_kwargs['ls'] = ls if linestyle is None else linestyle

        ax.plot(cc, data * norm, **plot_kwargs)

    ax.legend()
    ax.set_xlabel(f'concentration ({concentration_unit})')
    ax.set_ylabel(f'{ppty.quantity.capitalize()} {ppty.unit}')


for ppty in solution_properties:
    fig, ax = plt.subplots()
    fig.suptitle('Solutions')
    for solute in ppty.solutes:
        kwargs = {'ctype': 'm', 'ax': ax, 'norm': 1}
        plot_all_sources_conc(ppty, solute, **kwargs)
        if ppty.quantity == 'electrical conductivity':
            plot_all_sources_conc(ppty, solute, T=0, linestyle=':', **kwargs)
            plot_all_sources_conc(ppty, solute, T=50, linestyle='--', **kwargs)
    fig.tight_layout()


# ---------------------------- Solubility diagrams ---------------------------

fig, ax = plt.subplots()


# KCl and LiCl --------------

T_KCl = np.linspace(0, 50)
T_LiCl = np.linspace(10, 25)

sol_KCl = solubility('KCl', T=T_KCl)
sol_LiCl = solubility('LiCl', T=T_LiCl)

ax.plot(T_KCl, sol_KCl, '-', c=COLORS['KCl'], label='KCl')
ax.plot(T_LiCl, sol_LiCl, '-', c=COLORS['LiCl'], label='LiCl')


# NaCl --------------

T_hydrohalite = np.linspace(-10, 0.1)
T_halite = np.linspace(0.1, 50)

sol_hydrohalite = solubility('NaCl,2H2O', T=T_hydrohalite)
sol_halite = solubility('NaCl', T=T_halite)

c = COLORS['NaCl']

ax.plot(T_hydrohalite, sol_hydrohalite, '--', c=c, label='NaCl,2H2O (hydrohalite)')
ax.plot(T_halite, sol_halite, '-', c=c, label='NaCl (halite)')


# Na2SO4 ----------

T_mirabilite = np.linspace(0, 32.38)
T_thenardite = np.linspace(0, 50)

sol_mirabilite = solubility('Na2SO4,10H2O', T=T_mirabilite)
sol_thenardite = solubility('Na2SO4', T=T_thenardite)

c = COLORS['Na2SO4']

ax.plot(T_mirabilite, sol_mirabilite, '--', c=c, label='Na2SO4,10H2O (mirabilite)')
ax.plot(T_thenardite, sol_thenardite, '-', c=c, label='Na2SO4 (thenardite)')


# LiBr ------------

T3 = np.linspace(-25, 5.7)
T2 = np.linspace(5.7, 34.6)
T1 = np.linspace(34.6, 50)

sol1 = solubility('LiBr,H2O', T=T1)
sol2 = solubility('LiBr,2H2O', T=T2)
sol3 = solubility('LiBr,3H2O', T=T3)

c = COLORS['LiBr']

ax.plot(T1, sol1, '--', c=c, label='LiBr,H2O')
ax.plot(T2, sol2, '-', c=c, label='LiBr,2H2O')
ax.plot(T3, sol3, ':', c=c, label='LiBr,3H2O')


# -----------------

ax.set_xlabel('T (°C)')
ax.set_ylabel('$m_\mathrm{sat}$ [mol/kg]')
ax.set_xlim(-10, 50)
ax.grid()
ax.legend()


# ================================ FINAL =====================================


fig_w.tight_layout()
plt.show()



"""Ionic strength and individual ionic concentrations etc.

SOURCES
-------
Personal calculations, and:
ionic strength expressed as molar fraction is used in Clegg et al. AST, 1997
ionic strength expressed as molality is more common, e.g. Pitzer 1973
"""

# Local imports
from ...format import format_input_type, check_solute
from ...constants import solute_list
from ...constants import dissociation_numbers, charge_numbers


# ========================== INDIVIDUAL ION QUANTITIES =======================


def ion_quantities(solute, **concentration):
    """Return quantities x, m, c but defined for each ion instead of considering
    the solute as a single species. Used in ionic strength calculations.

    ion_quantities('NaCl', x=0.1) returns (x1, x2) where x1 is the mole fraction
    of Na+, x2 is that of Cl-.

    In this situation, one considers that there are three components in solution
    i.e. the cation (x1), the anion (x2) and the solvent (xw), with
    xi = ni / (ni + nj + nw).

    For molalities or concentrations, things are easier because these quantities
    are just multiplied by the dissociation number when considering individual
    ions compared to the solute as a whole. They are calculated using e.g.
    ion_quantities('NaCl', m=5.3) or ion_quantities('NaCl', c=4.8)
    """
    if len(concentration) == 1:
        param, = concentration.keys()  # param is the chosen parameter ('x', 'm' or 'c')
        value, = concentration.values()  # corresponding value in the unit above
    else:
        raise ValueError('kwargs must have exactly one keyword argument for solute concentration.')

    value = format_input_type(value)  # allows for lists and tuples as inputs

    check_solute(solute, solute_list)
    n1, n2 = dissociation_numbers[solute]

    if param == 'x':
        x = value
        ntot = n1 + n2
        x_interm = x / (1 + x * (ntot - 1))  # just for calculation
        x1 = n1 * x_interm
        x2 = n2 * x_interm
        return x1, x2

    elif param in ['m', 'c']:  # in this case, things are simply additive
        return n1 * value, n2 * value

    else:
        raise ValueError(f"quantity {param} not allowed, should be 'x', 'm' or 'c'.")


# =============================== IONIC STRENGTH =============================


def ionic_strength(solute, **concentration):
    """Ionic strength in terms of mole fraction (x), molality (m) or molarity (c)

    ionic_strength('NaCl', x=0.1) returns the mole fraction ionic strength (Ix)
    ionic_strength('NaCl', m=1.2) returns the molal ionic strength
    ionic_strength('NaCl', c=5.3) returns the molar ionic strength

    Note: output units is different for each case (same as input parameter, e.g.
    mol / m^3 for 'c').
    """
    check_solute(solute, solute_list)
    z1, z2 = charge_numbers[solute]
    y1, y2 = ion_quantities(solute, **concentration)
    I_strength = 0.5 * (y1 * z1 ** 2 + y2 * z2 ** 2)
    return I_strength



"""Activity coefficients of solutions"""

from .LiBr import ActivityCoefficientFormulas_LiBr
from .LiCl import ActivityCoefficientFormulas_LiCl
from .KCl import ActivityCoefficientFormulas_KCl
from .Na2SO4 import ActivityCoefficientFormulas_Na2SO4
from .NaCl import ActivityCoefficientFormulas_NaCl

ActivityCoefficientFormulas = (
    ActivityCoefficientFormulas_LiBr +
    ActivityCoefficientFormulas_LiCl +
    ActivityCoefficientFormulas_KCl +
    ActivityCoefficientFormulas_Na2SO4 +
    ActivityCoefficientFormulas_NaCl
)



"""Function to calculate the desnity of ambient water as a function of
temperature using IAPWS recommended equations or others.

Sources
-------

--- 'IAPWS'
    Patek et al.
    "Reference Correlations for Thermophysical Properties of Liquid Water
    at 0.1 MPa"
    J. Phys. Chem. Ref. Data
    (2009)

--- 'Kell'
    Kell
    "Density, Thermal Expansivity, and Compressibility of Liquid Water from
    0° to 150°C: Correlations and Tables for Atmospheric Pressure and
    Saturation Reviewed and Expressed on 1968 Temperature Scale"
    J. Chem. Eng. Data.
    (1975)
"""

from ...constants import Patm
from ..general import WaterFormula


class DensityAtm_IAPWS(WaterFormula):

    source ='IAPWS'
    temperature_unit = 'K'
    temperature_range = (253.15, 383.15)
    default = True

    def calculate(self, T):
        """Saturated liquid water density according to Patek 2009

        Input
        -----
        Temperature in K

        Output
        ------
        Density in kg/m^3

        Reference
        ---------
        Patek et al., "Reference Correlations for Thermophysical Properties of
        Liquid Water at 0.1 MPa", J. Phys. Chem. Ref. Data, 2009.
        (recommended by IAPWS)

        Notes
        -----
        - Valid between 253.15 K and 283.15 K
        """
        p0 = 0.1e6  # Pa
        R = 461.51805  # J.kg^−1.K^−1
        Tr = 10  # K
        Ta = 593  # K
        Tb = 232  # K

        alpha = Tr / (Ta - T)
        beta = Tr / (T - Tb)

        a5 = 1.93763157e-2; a6 = 6.74458446e3; a7 = -2.22521604e5; a8 = 1.00231247e8
        a9 = -1.63552118e9; a10 = 8.32299658e9; a11 = -7.5245878e-6; a12 = -1.3767418e-2
        a13 = 1.0627293e1; a14 = -2.0457795e2; a15 = 1.2037414e3

        n6 = 4; n7 = 5; n8 = 7; n9 = 8; n10 = 9
        n11 = 1; n12 = 3; n13 = 5; n14 = 6; n15 = 7

        b5 = 5.78545292e-3; b6 = -1.53195665e-2; b7 = 3.11337859e-2; b8 = -4.23546241e-2
        b9 = 3.38713507e-2; b10 = -1.19946761e-2; b11 = -3.1091470e-6; b12 = 2.8964919e-5
        b13 = -1.3112763e-4; b14 = 3.0410453e-4; b15 = -3.9034594e-4; b16 = 2.3403117e-4
        b17 = -4.8510101e-5

        m5 = 1; m6 = 2; m7 = 3; m8 = 4; m9 = 5; m10 = 6; m11 = 1
        m12 = 3; m13 = 4; m14 = 5; m15 = 6; m16 = 7; m17 = 9

        v0 = R * Tr / p0 * (a5 + a6 * alpha**n6 + a7 * alpha**n7 + a8 * alpha**n8
                            + a9 * alpha**n9 + a10 * alpha**n10 + b5 * beta**m5
                            + b6 * beta**m6 + b7 * beta**m7 + b8 * beta**m8
                            + b9 * beta**m9 + b10 * beta**m10)

        v_p0 = R * Tr / p0**2 * (a11 * alpha**n11 + a12 * alpha**n12 + a13 * alpha**n13
                                + a14 * alpha**n14 + a15 * alpha**n15 + b11 * beta**m11
                                + b12 * beta**m12 + b13 * beta**m13 + b14 * beta**m14
                                + b15 * beta**m15 + b16 * beta**m16 + b17 * beta**m17)

        v_atm = v0 + v_p0 * (Patm - p0)

        rho = 1 / v_atm

        return rho



class DensityAtm_Kell(WaterFormula):

    source ='Kell'
    temperature_unit = 'C'
    temperature_range = (0, 150)

    def calculate(self, T):
        """Ambient water density according to Kell 1975

        Input
        -----
        Temperature in C

        Output
        ------
        Density in kg/m^3

        Reference
        ---------
        - Kell : "Density, Thermal Expansivity, and Compressibility of Liquid
        Water from 0° to 150°C: Correlations and Tables for Atmospheric Pressure
        and Saturation Reviewed and Expressed on 1968 Temperature Scale",
        J. Chem. Eng. Data. 1975

        Notes
        -----
        - Used by Clegg2011
        - Valid between 0°C and 150°C
        """
        rho = 999.83952 + 16.945176 * T - 7.9870401e-3 * T**2 - 46.170461e-6 * T**3 + 105.56302e-9 * T**4 - 280.54253e-12 * T**5
        rho = rho / (1 + 16.879850e-3 * T)
        return rho


# ========================== WRAP-UP OF FORMULAS =============================

DensityAtmFormulas = (
    DensityAtm_IAPWS,
    DensityAtm_Kell,
)



"""Functions to calculate the surface tension of water as a function of T.

Sources
-------
--- 'IAPWS':
    Release on Surface Tension of Ordinary Water Substance, IAPWS, London
    (September 1994)
"""

from ...constants import Tc
from ..general import WaterFormula


class SurfaceTension_IAPWS(WaterFormula):
    """Water surface tension as a function of temperature, from IAPWS"""

    source ='IAPWS'
    temperature_unit = 'K'
    temperature_range = (248.15, Tc)   # Extrapolated range included
    default = True

    coeffs = {
        'B': 235.8e-3,
        'b': -0.625,
        'mu': 1.256,
    }

    def calculate(self, T):
        """Formula of surface tension according to IAPWS.

        Input
        -----
        Temperature in K

        Output
        ------
        Surface tension in N/m

        Reference
        ---------
        IAPWS, Revised Release on Surface Tension of Ordinary Water Substance,
        IAPWS, Moscow, Russia, June 2014.

        Notes
        -----
        - Used by Conde2004 and Dutcher2010 for the surface tension of solutions.
        - Valid between triple point (0.01°C) and critical temperature 647.096K.
        It also provides reasonably accurate values when extrapolated into the
        supercooled region, to temperatures as low as -25°C.
        """
        tau = (1 - T / Tc)
        B, b, mu = self.coeffs.values()
        sigma = B * tau ** mu * (1 + b * tau)  # in N / m
        return sigma


# ========================== WRAP-UP OF FORMULAS =============================

SurfaceTensionFormulas = (
    SurfaceTension_IAPWS,
)



"""MISC. formula for surface tension of solutions"""

import numpy as np

from ...water.surface_tension import SurfaceTension_IAPWS


def sigma_iapws(T):
    """T in K, IAPWS formula"""
    surface_tension_iapws = SurfaceTension_IAPWS()
    return surface_tension_iapws.calculate(T=T)



"""Module with values of constants useful for solutions.

Note: dictionaries written in alphabetical order.

CONTENTS
--------
General properties:
    - Na (float): Avogadro's constant
    - R (float): molar gas constant
    - Patm (float): atmospheric pressure in Pa
Water properties:
    - Mw (float): molar mass of water in kg / mol
    - Tc (float): critical temperature in K
    - Pc (float): critical pressure in Pa
    - rhoc (float): critical density in kg / m^3
Ion properties in dictionaries (keys : ion name, e.g. 'Cl')
    - weight_cations: molecular weight of cations in Daltons
    - weight_anions: molecular weight of anions in Daltons
Solute properties as dictionaries (keys: solute name, e.g. 'NaCl')
    - dissociation_numbers: tuple of  number of cations and anions released
    - charge_numbers: tuple of unit charges of cation and anion
Solute properties as functions:
    - molar_mass(solute). Input: solute name (e.g. 'NaCl'), output M in kg / mol

SOURCES
-------
CRC Handbook of Physics and Chemistry:
    - Physical Constants of Inorganic Compounds
    https://hbcp.chemnetbase.com/faces/documents/04_02/04_02_0001.xhtml
    - Thermodynamic Properties of Aqueous Ions
    http://hbcponline.com/faces/documents/05_04/05_04_0001.xhtml
    - Recommended values of the fundamental physical constants
    http://hbcponline.com/faces/documents/01_01/01_01_0001.xhtml
    - Fixed-point properties of H20 and D20
    http://hbcponline.com/faces/documents/06_04/06_04_0001.xhtml

IAPWS, Release on Surface Tension of Ordinary Water Substance
IAPWS, London, September 1994.

"""

# ============================= GENERAL CONSTANTS ============================

Na = 6.02214085774e23  # Avogadro's constant
R = 8.314459848  # molar gas constant in J/(mol.K)
Patm = 101325  # atmospheric pressure in Pa

# ============================== WATER PROPERTIES ============================

Mw = 18.015268e-3  # molar mass in kg / mol
Tc = 647.096  # critical temperature in K (IAPWS 2014)
Pc = 22.064e6  # critical pressure in Pa (CRC Handbook & IAPWS)
rhoc = 322    # critical density in kg/m^3 (CRC Handbook & IAPWS)

# =========================== SOLUTE/IONS PROPERTIES =========================

solute_list = [
    'AlCl3',
    'CaCl2',
    'K2SO4',
    'KCl',
    'KI',
    'KNO3',
    'LiBr',
    'LiCl',
    'MgCl2',
    'MgSO4',
    'Na2SO4',
    'NaCl',
    'NaNO3',
]

# Individual ion molecular weights in Daltons --------------------------------
weight_cations = {
    'Al': 26.982,
    'Ca': 40.078,
    'K': 39.098,
    'Li': 6.94,
    'Mg': 24.305,
    'Na': 22.99,
}

weight_anions = {
    'Br': 79.904,
    'Cl': 35.453,
    'I': 126.904,
    'NO3': 62.005,
    'SO3': 80.063,
    'SO4': 96.063,
}

# Number of ions per solute molecule after dissociation, for cation and anion
dissociation_numbers = {
    'AlCl3': (1, 3),
    'CaCl2': (1, 2),
    'K2SO4': (2, 1),
    'KCl': (1, 1),
    'KI': (1, 1),
    'KNO3': (1, 1),
    'LiBr': (1, 1),
    'LiCl': (1, 1),
    'MgCl2': (1, 2),
    'MgSO4': (1, 1),
    'Na2SO4': (2, 1),
    'NaCl': (1, 1),
    'NaNO3': (1, 1),
}

# Unit charges of cation and anion in the solute -----------------------------
charge_numbers = {
    'AlCl3': (3, 1),
    'CaCl2': (2, 1),
    'K2SO4': (1, 2),
    'KCl': (1, 1),
    'KI': (1, 1),
    'KNO3': (1, 1),
    'LiBr': (1, 1),
    'LiCl': (1, 1),
    'MgCl2': (2, 1),
    'MgSO4': (2, 2),
    'Na2SO4': (1, 2),
    'NaCl': (1, 1),
    'NaNO3': (1, 1),
}

# Individual ions composing the molecule -------------------------------------
individual_ions = {
    'AlCl3': ('Al', 'Cl'),
    'CaCl2': ('Ca', 'Cl'),
    'K2SO4': ('K', 'SO4'),
    'KCl': ('K', 'Cl'),
    'KI': ('K', 'I'),
    'KNO3': ('K', 'NO3'),
    'LiBr': ('Li', 'Br'),
    'LiCl': ('Li', 'Cl'),
    'MgCl2': ('Mg', 'Cl'),
    'MgSO4': ('Mg', 'SO4'),
    'Na2SO4': ('Na', 'SO4'),
    'NaCl': ('Na', 'Cl'),
    'NaNO3': ('Na', 'NO3'),
}

# calculation of molar mass from the molecular weights -----------------------


def molar_mass(solute):
    """Return molar mass of solute compound in kg / mol."""
    try:
        cation, anion = individual_ions[solute]
    except KeyError:
        raise KeyError(f'{solute} molecular/molar mass data not available')

    m1, m2 = weight_cations[cation], weight_anions[anion]
    nu1, nu2 = dissociation_numbers[solute]
    mtot = nu1 * m1 + nu2 * m2

    return mtot * 1e-3



"""Gathers the formulas for the density of KCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns two parameters:
- rho0, density of pure water in kg / m^3
- rho, density of solution in kg / m^3
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Al Ghafri et al., Densities of Aqueous MgCl2(aq), CaCl2 (aq), KI(aq),
NaCl(aq), KCl(aq), AlCl3(aq), and (0.964 NaCl + 0.136 KCl)(aq) at
Temperatures Between (283 and 472) K, Pressures up to 68.5 MPa, and
Molalities up to 6 mol·kg -1.
Journal of Chemical & Engineering Data 57, 1288-1304 (2012).

- Krumgalz, B. S., Pogorelsky, R. & Pitzer, K. S.
  Volumetric Properties of Single Aqueous Electrolytes from Zero to Saturation
  Concentration at 298.15 °K Represented by Pitzer's Ion-Interaction Equations.
  Journal of Physical and Chemical Reference Data 25, 663-689 (1996).
"""

from .al_ghafri import Density_KCl_AlGhafri_Base
from .krumgalz import Density_KCl_Krumgalz_Base


class Density_KCl_AlGhafri(Density_KCl_AlGhafri_Base):
    """Already defined in Al Ghafri module"""
    default = True

class Density_KCl_Krumgalz(Density_KCl_Krumgalz_Base):
    """Already defined in Krumgalz module and not default here"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

Density_KCl_Formulas = (
    Density_KCl_AlGhafri,
    Density_KCl_Krumgalz,
)


"""Index of refraction of solutions"""

from .CaCl2 import RefractiveIndexFormulas_CaCl2
from .KCl import RefractiveIndexFormulas_KCl
from .NaCl import RefractiveIndexFormulas_NaCl

RefractiveIndexFormulas = (
    RefractiveIndexFormulas_CaCl2 +
    RefractiveIndexFormulas_KCl +
    RefractiveIndexFormulas_NaCl
)



"""Activity of solutions according to Tang

NOTE: Very similar in structure to activity_coefficient.tang

Source
------
- Tang, I. N., Munkelwitz, H. R. & Wang, N.
  Water activity measurements with single suspended droplets:
  The NaCl-H2O and KCl-H2O systems.
  Journal of Colloid and Interface Science 114, 409-415 (1986).
  Valid at 25°C and for solutions of molality up to ~13 mol/kg
"""

import numpy as np

from ....constants import Mw, charge_numbers, dissociation_numbers
from ...general import SolutionFormula


def aw_extended_debye_huckel(m, T, solute, coeffs):
    """Mix of Hamer & Wu 1972 and Tang, Munkelwitz and Wang 1986.

    Used for NaCl and KCl
    """
    z1, z2 = charge_numbers[solute]
    nu = sum(dissociation_numbers[solute])

    A, B, C, D, E, beta = coeffs

    b = 1 + B * np.sqrt(m)

    term1 = (z1 * z2 * A / (B**3 * m)) * (b - 4.60517 * np.log10(b) - 1 / b)
    term2 = - (beta * m / 2) - (2 / 3 * C * m**2) - (3 / 4 * D * m**3) - (4 / 5 * E * m**4)

    phi =  1 - 2.302585 * (term1 + term2)  # osmotic coefficient

    return np.exp(-nu * Mw * phi * m)


class WaterActivity_Tang_Base(SolutionFormula):

    source ='Tang'

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'm'

    with_water_reference = False

    def calculate(self, m, T):
        return aw_extended_debye_huckel(
            m=m,
            T=T,
            solute=self.solute,
            coeffs=self.coeffs.values(),
        )


class WaterActivity_NaCl_Tang_Base(WaterActivity_Tang_Base):

    solute = 'NaCl'
    concentration_range = (1e-9, 14)

    coeffs = {
        'A': 0.5108,
        'B': 1.37,
        'C': 4.803e-3,
        'D': -2.736e-4,
        'E': 0,
        'beta': 2.796e-2,
    }


class WaterActivity_KCl_Tang_Base(WaterActivity_Tang_Base):

    solute = 'KCl'
    concentration_range = (1e-9, 13)

    coeffs = {
        'A': 0.5108,
        'B': 1.35,
        'C': 7.625e-3,
        'D': -7.892e-4,
        'E': 2.492e-5,
        'beta': -9.842e-3,
    }




"""Various tools to calculate solution densities using Clegg formulae."""


import numpy as np

from ....constants import dissociation_numbers, Mw, molar_mass, charge_numbers
from ...water.density_atm import DensityAtm_IAPWS
from ..ionic import ionic_strength
from ..pitzer import PitzerVolumetric


Av = 1.8305  # From Archer & Wang 1990, in cm^3 kg^(1/2) mol^(-3/2)
Avx = Av * (1 / Mw) ** (1/2)  # The 1000 factor from Clegg disappears here due to Mw being in SI
rho_bold = 13  # bold rho parameter, not a density


# ----------------------------------------------------------------------------
# =================== Formulas for NaCl and Na2SO4 at 25°C ===================
# ----------------------------------------------------------------------------


params_NaCl = {
    'c1': 0.21532202E+01,  # This coefficient is WRONG in the paper's SI (indicated as E+02)
    'c2': 0,
    'c3': -0.66928951E+01,
    'c4': 0.15530704E+03,
    'c5': 0.12662763E+03,
    'c6': -0.26841488E+03,  # COEFF w**(1.75) instead of w**(2.5)
    'c7': 0,
    'c8': 0,
    'V_phi_infty': 16.62,
}

params_Na2SO4 = {
    'c1': -0.99167549E+01,
    'c2': 0.95929970E+02,
    'c3': -0.11648247E+03,
    'c4': 0.18800889E+03,
    'c5': -0.21207426E+03,
    'c6': 0,
    'c7': 0.25016809E+03,
    'c8': -0.15977556E+03,
    'V_phi_infty': 11.78,
}

params_all = {
    'NaCl': params_NaCl,
    'Na2SO4': params_Na2SO4,
}


def apparent_molar_volume_25(w, solute='NaCl'):
    """Apparent molar volume in m^3 / mol at 25°C"""

    powers = {'c1': 0.5,
              'c2': 0.75,
              'c3': 1,
              'c4': 1.5,
              'c5': 2,
              'c6': 2.5,
              'c7': 3,
              'c8': 3.5}

    params = params_all[solute]

    if solute == 'NaCl':
        powers['c6'] = 1.75

    z_m, z_x = charge_numbers[solute]
    nu_m, nu_x = dissociation_numbers[solute]
    nu_mx = nu_m + nu_x

    # Note: I use for now the 2 lines below instead of convert() because of
    # current unavailability of the convert() function [broken temporarily]
    M = molar_mass(solute)
    x = (w / M) / (w / M + (1 - w) / Mw)
    Ix = ionic_strength(solute, x=x)

    V_phi = params['V_phi_infty']
    V_phi += nu_mx * z_m * z_x * Avx / (2 * rho_bold) * np.log(1 + rho_bold * np.sqrt(Ix))

    for coeff, power in powers.items():
        V_phi += params[coeff] * w ** power

    return V_phi * 1e-6  # conversion cm^3 to m^3


def density_25(w, solute='NaCl'):
    """Density at 25°C in kg / m^3"""

    M = molar_mass(solute)

    # Note: I use for now the line below instead of convert() because of
    # current unavailability of the convert() function [broken temporarily]
    m = w / ((1 - w) * M)

    density_atm = DensityAtm_IAPWS()
    rho_w = density_atm.calculate(T=298.15)
    V_phi = apparent_molar_volume_25(w, solute=solute)

    return rho_w * (1 + m * M) / (1 + m * rho_w * V_phi)


# ----------------------------------------------------------------------------
# =================== Formula for NaCl at all temperatures ===================
# ---------------- (and for concentrations above saturation) -----------------
# ----------------------------------------------------------------------------


def q1(w):
    """First temperature coefficient for NaCl"""
    return -5.2816566e-4 - 2.443434e-5 * w


def q2(w):
    """Second temperature coefficient for NaCl"""
    return -2.421396e-6 + 2.421396e-6 * w


def density_NaCl(w, T):
    """T in Kelvins (K)"""
    Tr = 298.15
    rho = density_25(w, solute='NaCl')

    # lines below are equivalent to this version displayed in the paper:
    # rho += (T - Tr) * (q1(w) - q2(w) * Tr) * 1e3
    # rho += q2(w) / 2 * (T**2 - Tr**2) * 1e3

    rho += q1(w) * (T - Tr) * 1e3  # 1e3 for conversion g/cm^3 --> kg / m^3
    rho += q2(w) * (T - Tr)**2 / 2 * 1e3

    return rho


# ----------------------------------------------------------------------------
# ================== Formula for N2SO4 at all temperatures ===================
# ----------------------------------------------------------------------------


# High concentration ---------------------------------------------------------


def Q1(nu):
    """First temperature coefficient for Na2SO4"""
    return -5.399e-4 + 0.80541318e-3 * np.sqrt(nu) - 0.00041107238 * nu**2


def Q2(nu):
    """Second temperature coefficient for Na2SO4"""
    return -0.27716993e-8 * nu + 1.3723983e-9 * nu**4


def density_Na2SO4_high_conc(w, T):
    """Valid above weight fraction of 22%. T in Kelvins (K)"""
    Tr = 298.15
    rho = density_25(w, solute='Na2SO4')

    nu = (1 - w) / 0.78

    rho += Q1(nu) * (T - Tr) * 1e3
    rho += Q2(nu) * (T**3 - Tr**3) * 1e3

    return rho


# Low concentration ----------------------------------------------------------
# NOTE -- There seems to be a problem with the low-concentration formulas
# below, because the values for the apparent molar volumes and densities
# do not match the tables from the Supplemental Information in Clegg's paper.
# (also, it is supposed to have a continuity with the high-concentration
# formulas at w=0.22 -- the "reference concentratinon" and it is not the case)


beta0_coeffs = 0.0024497654, 0, -0.48224610e-3, 0.11372104e-5
beta1_coeffs = -1.0086131, 28.944503, 0.17371696, -0.26199156e-3
C0_coeffs = 0.59267496e-5, -0.0022005223, 0, 0
alpha = 1.4
Ratm = 82.05746  # idel gas constants when atm is used instead of MPa


def v_phi_infty(T):
    """Apparent molar volume at infinite dilution (T in K)"""
    return 28746.8530 + 6.802415 * T - 4886.246974 * np.log(T) - 8.715969e5 / T


def _calculate_parameter(T, *coeffs):
    """Temperature correlations for parameters beta0, beta1, C0.

    T in K
    """
    c0, c1, c2, c3 = coeffs
    return c0 + c1 / T + c2 * np.log(T) + c3 * T

def density_Na2SO4_low_conc(m, T):
    """Density at 25°C in kg / m^3. T in Kelvins (K)"""
    beta0 = _calculate_parameter(T, *beta0_coeffs)
    beta1 = _calculate_parameter(T, *beta1_coeffs)
    Cv = _calculate_parameter(T, *C0_coeffs)
    v0 = v_phi_infty(T)

    pitz = PitzerVolumetric(
        solute='Na2SO4',
        T=T,
        R=Ratm,
        A_v=Av,
        C_v=Cv,
        v_0=v0,
        beta0=beta0,
        beta1=beta1,
        alpha1=alpha,
    )

    v_phi = pitz.apparent_molar_volume(m=m) * 1e-6  # conversion cm^3 --> m^3

    Ms = molar_mass('Na2SO4')

    density_atm = DensityAtm_IAPWS()
    rho_w = density_atm.calculate(T)

    return rho_w * (1 + m * Ms) / (1 + m * rho_w * v_phi)



"""Gathers the formulas for the surface tension of NaCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns two parameters
- sigma0, sigma surface tensions in N/m of pure water and solution
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Dutcher: Dutcher, C. S., Wexler, A. S. & Clegg, S. L.
  Surface Tensions of Inorganic Multicomponent Aqueous Electrolyte Solutions and Melts.
  J. Phys. Chem. A 114, 12216-12230 (2010).

- Talreja-Muthreja, T., Linnow, K., Enke, D. & Steiger
  M. Deliquescence of NaCl Confined in Nanoporous Silica.
  Langmuir 38, 10963-10974 (2022).
"""

# TODO: add data from Ali 2006

from ...general import SolutionFormula
from .dutcher import SufaceTension_NaCl_Dutcher_Base
from .misc import sigma_iapws


class SurfaceTension_NaCl_Dutcher(SufaceTension_NaCl_Dutcher_Base):
    """Already defined in dutcher module"""
    default = True


class SurfaceTension_NaCl_Steiger(SolutionFormula):

    source ='Steiger'
    solute = 'NaCl'

    temperature_unit = 'C'
    temperature_range = (-10, 50)

    concentration_unit = 'm'
    concentration_range = (0, 7)     # approx (up to saturation)

    with_water_reference = True

    def calculate(self, m, T):
        """Surface tension calculated from Talreja-Muthreja et al. 2022
        Input: molality m, temperature T in Celsius."""
        sigma_w = sigma_iapws(T + 273.15)
        sigma = sigma_w + 0.00166 * m
        return sigma_w, sigma


# ========================== WRAP-UP OF FORMULAS =============================

SurfaceTensionFormulas_NaCl = (
    SurfaceTension_NaCl_Dutcher,
    SurfaceTension_NaCl_Steiger,
)


"""Inverse functions for solutions"""

# Non standard imports
import numpy as np
from pynverse import inversefunc

# local imports
# (DO NOT import from . because it will create circular import problems)
from .properties import water_activity
from .convert import convert

from ..format import format_output_type


# ======================== ACTIVITY TO CONCENTRATION =========================


def aw_to_conc(a, out='w', solute='NaCl', T=25, unit='C', source=None):
    """Calculate concentration needed to achieve a certain water activity.

    Parameters
    ----------
    - a is the water activity,
    - out: any concentration unit manageable by convert()
    - solute (default NaCl) is the solute of interest
    - T: temperature
    - unit: temperature unit ('C' or 'K')
    - source: if None, use default source.

    Output
    ------
    Concentration using the units asked for with the "out" parameter.

    Examples
    --------
    aw_to_conc(0.39)
    >>> 0.4902761745068064  # in terms of weight fraction
    aw_to_conc([0.39, 0.75], out='m')
    >>> array([16.45785963,  6.21127029])  # in terms of molality
    aw_to_conc(0.11, 'r', 'LiCl', T=50)
    >>> 0.9167650291014361  # in terms of mass ratio, for LiCl, at 50°C

    Note: part of the structure of this function resembles that of
    general.calculation(), so see if there is a way to avoid redundancy
    """
    formula = water_activity.get_formula(solute=solute, source=source)
    cunit = formula.concentration_unit
    cmin, cmax = formula.concentration_range

    def activity(conc):
        return water_activity(
            solute=solute,
            T=T,
            unit=unit,
            source=source,
            **{cunit: conc},
        )

    # If activity goes through a min, replace cmax to avoid problems w/ invert
    cc = np.linspace(cmin, cmax, num=200)
    aa = activity(cc)
    imin = np.argmin(aa)
    cmax = cc[imin] # will be equal to cmax if function only decreases
    amin = aa[imin]

    concentration = inversefunc(activity, domain=[cmin, cmax])

    try:
        conc = concentration(a)
    except ValueError:
        src = water_activity.get_source(solute=solute, source=source)
        msg = (
            f"Water activity a={a} lower than minimum achievable with {src}'s'"
            f"formula (a={amin:.3f})"
        )
        raise ValueError(msg)
    else:
        c = convert(
            value=conc,
            unit1=cunit,
            unit2=out,
            solute=solute,
            T=T,
            unit=unit,
        )
        return format_output_type(c)



"""Gathers the formulas for the electrical conductivity of KCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns one parameter
- sigma, electrical conductivity, in S/m
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- McKee, C. B.
  An Accurate Equation for the Electrolytic Conductivity
  of Potassium Chloride Solutions.
  J Solution Chem 38, 1155-1172 (2009).
"""

import numpy as np

from ...general import SolutionFormula


# ============================== FORMULAS ====================================

# ----------------------------------- MC25 -----------------------------------


def mc25_flas(m):
    """Conductivity component FLAS from McKee 2009"""
    a = 111.732
    b = -6.01942e-2
    c = -2.03830e-3
    return a * np.exp(b * m + c * m**2)


def mc25_fmid(m):
    """Conductivity component FMID from McKee 2009"""
    a = 34.5286
    b = -2.25008
    return a * np.exp(b * m**(1/2))


def mc25_fini(m):
    """Conductivity component FINI from McKee 2009"""
    a = 1.45667
    b = -19.5946
    return a * np.exp(b * m)


def mc25_finii(m):
    """Conductivity component FINII from McKee 2009"""
    a = 0.960521
    b = -96.4192
    return a * np.exp(b * m)


def mc25_finiii(m):
    """Conductivity component FINIII from McKee 2009"""
    a = 0.587725
    b = -570.452
    return a * np.exp(b * m)


def mc25(m):
    """Conductivity, Total"""
    return mc25_flas(m) + mc25_fmid(m) + mc25_fini(m) + mc25_finii(m) + mc25_finiii(m)


# ------------------------------- MC50 / MC25 --------------------------------


fcon0 = 0.321662


def sol50re_flas(m):
    """Terms for MC50 / MC25, FLAS"""
    a = 0.165123
    b = -0.265938
    return a * np.exp(b * m)


def sol50re_fmid(m):
    """Terms for MC50 / MC25, FMID"""
    a = 1.57042e-2
    b = -2.16681
    return a * np.exp(b * m)


def sol50re_fini(m):
    """Terms for MC50 / MC25, FINI"""
    a = 8.26838e-3
    b = -26.7895
    return a * np.exp(b * m)


def sol50re_finii(m):
    """Terms for MC50 / MC25, FINII"""
    a = 2.99403e-3
    b = -525.162
    return a * np.exp(b * m)


def sol50re(m):
    """SOL50RE term"""
    return sol50re_flas(m) + sol50re_fmid(m) + sol50re_fini(m) + sol50re_finii(m)


def mc50_mc25(m):
    return 1 + fcon0 + sol50re(m)


# ------------------------ Temperature extrapolation -------------------------


def alpha_v(T):
    """ALPHA_V factor, T is temperature in deg. C"""
    b = -6.35249e-4
    c = 4.60006e-4
    d = -4.97223e-3
    e = -6.85638e-4
    return b + c * np.exp(d * T + e * T**2)


def alpha(T):
    """ALPHA factor, T is temperature in deg. C"""
    T1 = T - 25
    T2 = T - 50
    a = 4e-2
    return T1 * (a + T2 * alpha_v(T))


def beta_a_v(T):
    """BETAa_V factor, T is temperature in deg. C"""
    b = 1.21005e-3
    c = 6.64169e-4
    d = -2.64072e-3
    e = -7.78958e-4
    return b - c * np.exp(d * T + e * T**2)


def beta_a(T):
    """BETAa factor, T is temperature in deg. C"""

    T1 = T - 25
    T2 = T - 50
    a = 4e-2
    return T1 * (a + T2 * beta_a_v(T))


def beta_b_v(T):
    """BETAb_V factor, T is temperature in deg. C"""
    b = 4.44279e-4
    c = 3.35363e-4
    d = -3.04164e-3
    e = -6.99205e-4
    return b - c * np.exp(d * T + e * T**2)


def beta_b(T):
    """BETAb factor, T is temperature in deg. C"""
    T1 = T - 25
    T2 = T - 50
    return T1 * T2 * beta_b_v(T)


# ------------------------------- ALL TOGETHER -------------------------------

def mc_mc25(m, T):
    """MC / MC25"""
    return 1 + alpha(T) * fcon0 + (beta_a(T) + beta_b(T) * m) * sol50re(m)


def conductivity_KCl_McKee(m, T=25):
    return 0.1 * m * mc25(m) * mc_mc25(m, T)



class ElectricalConductivity_KCl_McKee(SolutionFormula):

    source ='McKee'
    solute = 'KCl'

    temperature_unit = 'C'
    temperature_range = (0, 55)

    concentration_unit = 'm'
    concentration_range = (0, 5)

    default = True
    with_water_reference = False

    def calculate(self, m, T):
        return conductivity_KCl_McKee(m=m, T=T)


# ========================== WRAP-UP OF FORMULAS =============================

ElectricalConductivity_KCl_Formulas = (
    ElectricalConductivity_KCl_McKee,
)



"""Duvall et al. formulas for solubility of solutions (LiBr).

Source
------
Duvall, K. N., Dirksen, J. A. & Ring, T. A.
Ostwald-Meyers Metastable Region in LiBr Crystallization—Comparison of
Measurements with Predictions.
Journal of Colloid and Interface Science 239, 391-398
(2001)

Note: the paper has sign errors that I have corrected
(see comments in coefficients below)
"""

import numpy as np

from ...general import SaturatedSolutionFormula


class Solubility_Duvall_Base(SaturatedSolutionFormula):

    source = 'Duvall'
    concentration_unit = 'x'
    temperature_unit = 'K'

    def calculate(self, T):
        """Returns mole fraction of saturated solution"""
        A, B, C = self.coeffs
        ln_x = A + B / T + C * np.log(T)
        return np.exp(ln_x)


# =============================== Steiger 2008 ===============================


class Solubility_LiBr_1H2O_Duvall_Base(Solubility_Duvall_Base):
    crystal = 'LiBr,H2O'
    temperature_range = tuple(273.15 + T for T in (34.6, 100))
    coeffs = -43.71368, 1858.62739, 6.366771  # sign error in paper last term


class Solubility_LiBr_2H2O_Duvall_Base(Solubility_Duvall_Base):
    crystal = 'LiBr,2H2O'
    temperature_range = tuple(273.15 + T for T in (5.7, 34.6))
    coeffs = -146.11879, 5671.38048, 22.078296


class Solubility_LiBr_3H2O_Duvall_Base(Solubility_Duvall_Base):
    crystal = 'LiBr,3H2O'
    temperature_range = tuple(273.15 + T for T in (-25, 5.7))  # lower bound approx
    coeffs = 377.41598, -16077.58765, -57.048733  # sign error in paper 2d term



"""Gathers the formulas for the activity coefficients of NaCl solutions.

Sources
-------

- Steiger, M.,
  Crystal growth in porous materials—I:
  The crystallization pressure of large crystals.
  Journal of Crystal Growth 282, 455-469 (2005).
  Valid at 25°C and up to 13.5 mol/kg

- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

(some info of domain of validity of expressions in the following paper:)
Dorn, J. & Steiger, M. Measurement and Calculation of Solubilities in the
Ternary System NaCH 3 COO + NaCl + H 2 O from 278 K to 323 K.
J. Chem. Eng. Data 52, 1784-1790 (2007).)

- Tang, I. N., Munkelwitz, H. R. & Wang, N.
  Water activity measurements with single suspended droplets:
  The NaCl-H2O and KCl-H2O systems.
  Journal of Colloid and Interface Science 114, 409-415 (1986).
  Valid at 25°C and for solutions of molality up to ~13 mol/kg

- Pitzer, K. S. & Mayorga, G.
  Thermodynamics of electrolytes. II.
  Activity and osmotic coefficients for strong electrolytes with one or both
  ions univalent.
  J. Phys. Chem. 77, 2300-2308
  (1973)
"""

from .pitzer import ActivityCoefficient_NaCl_Pitzer_Base
from .steiger import ActivityCoefficient_NaCl_Steiger2005_Base
from .steiger import ActivityCoefficient_NaCl_Steiger2008_Base
from .tang import ActivityCoefficient_NaCl_Tang_Base


class ActivityCoefficient_NaCl_Tang(ActivityCoefficient_NaCl_Tang_Base):
    """Already defined in tang module"""
    pass


class ActivityCoefficient_NaCl_Pitzer(ActivityCoefficient_NaCl_Pitzer_Base):
    """Already defined in pitzer module"""
    pass

class ActivityCoefficient_NaCl_Steiger2005(ActivityCoefficient_NaCl_Steiger2005_Base):
    """Already defined in steiger module"""
    pass


class ActivityCoefficient_NaCl_Steiger2008(ActivityCoefficient_NaCl_Steiger2008_Base):
    """Already defined in steiger module"""
    default = True


# ============================= WRAP-UP FORMULAS =============================

ActivityCoefficientFormulas_NaCl = (
    ActivityCoefficient_NaCl_Steiger2008,
    ActivityCoefficient_NaCl_Steiger2005,
    ActivityCoefficient_NaCl_Tang,
    ActivityCoefficient_NaCl_Pitzer,
)


"""Gathers the formulas for the electrical conductivity of NaCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns one parameter
- sigma, electrical conductivity, in S/m
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Sinmyo, R. & Keppler
  H. Electrical conductivity of NaCl-bearing aqueous fluids to 600 °C and 1 GPa.
  Contrib Mineral Petrol 172, 4 (2017).
"""

import numpy as np

from ...general import SolutionFormula
from ...water.density_atm import DensityAtm_IAPWS


class ElectricalConductivity_NaCl_Sinmyo(SolutionFormula):
    """FOR NOW THIS DOES NOT WORK WELL"""

    source ='Sinmyo'
    solute = 'NaCl'

    temperature_unit = 'K'
    temperature_range = (100 + 273.15, 600 + 273.15)

    concentration_unit = 'w'
    concentration_range = (0, 0.06)

    default = True
    with_water_reference = False

    @staticmethod
    def lambda_0(T, rho):
        """Molar conductivity of NaCl in water at infinite dilution (S cm^2 mol^-1).

        Used in Sinmyo's equation for NaCl solution conductivity.
        """
        return 1573 - 1212 * rho + 537_062 / T - 208_122_721 / T**2

    def calculate(self, w, T):
        """w weight fraction, T temperature in K"""
        density_atm = DensityAtm_IAPWS()
        rho = density_atm.calculate(T=T) / 1000  # pure water density in g / cm^3
        log_sigma = -1.7060 - 93.78 / T + 0.8075 * np.log(w) + 3.0781 * np.log(rho) + np.log(self.lambda_0(T, rho))
        return np.exp(log_sigma)


# ========================== WRAP-UP OF FORMULAS =============================

ElectricalConductivity_NaCl_Formulas = (
    # ElectricalConductivity_NaCl_Sinmyo,  NOT INCLUDED BECAUSE FAILS FOR NOW
)



"""Krumgalz density formulas for different salts

Source
------
- Krumgalz, B. S., Pogorelsky, R. & Pitzer, K. S.
  Volumetric Properties of Single Aqueous Electrolytes from Zero to Saturation
  Concentration at 298.15 °K Represented by Pitzer's Ion-Interaction Equations.
  Journal of Physical and Chemical Reference Data 25, 663-689 (1996).
"""

import numpy as np

from ...general import SolutionFormula
from .misc import density_pitzer

class Density_Krumgalz_Base(SolutionFormula):

    source ='Krumgalz'

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'm'

    with_water_reference = True

    def calculate(self, m, T):
        return density_pitzer(m, solute=self.solute, source='Krumgalz')


class Density_CaCl2_Krumgalz_Base(Density_Krumgalz_Base):
    solute = 'CaCl2'
    concentration_range = (0, 7.7)


class Density_KCl_Krumgalz_Base(Density_Krumgalz_Base):
    solute = 'KCl'
    concentration_range = (0, 4.7)


class Density_KI_Krumgalz_Base(Density_Krumgalz_Base):
    solute = 'KI'
    concentration_range = (0, 8.6)


class Density_LiCl_Krumgalz_Base(Density_Krumgalz_Base):
    solute = 'LiCl'
    concentration_range = (0, 19.6)


class Density_MgCl2_Krumgalz_Base(Density_Krumgalz_Base):
    solute = 'MgCl2'
    concentration_range = (0, 5.8)


class Density_Na2SO4_Krumgalz_Base(Density_Krumgalz_Base):
    solute = 'Na2SO4'
    concentration_range = (0, 1.5)


class Density_NaCl_Krumgalz_Base(Density_Krumgalz_Base):
    solute = 'NaCl'
    concentration_range = (0, 14)





"""Gathers the formulas for the activity of LiBr solutions.

Sources
-------
- Pitzer, K. S. & Mayorga, G.
  Thermodynamics of electrolytes. II.
  Activity and osmotic coefficients for strong electrolytes with one or both
  ions univalent.
  J. Phys. Chem. 77, 2300-2308
  (1973)
"""

from .patil import WaterActivity_LiBr_Patil_Base
from .pitzer import WaterActivity_LiBr_Pitzer_Base


class WaterActivity_LiBr_Patil(WaterActivity_LiBr_Patil_Base):
    """Already defined in pitzer module"""
    default = True


class WaterActivity_LiBr_Pitzer(WaterActivity_LiBr_Pitzer_Base):
    """Already defined in pitzer module"""
    pass



# ========================== WRAP-UP OF FORMULAS =============================

WaterActivityFormulas_LiBr = (
    WaterActivity_LiBr_Patil,
    WaterActivity_LiBr_Pitzer,
)



"""Activity coefficients of solutions according to Tang

NOTE: Very similar in structure to water_activity.tang

Source
------
- Tang, I. N., Munkelwitz, H. R. & Wang, N.
  Water activity measurements with single suspended droplets:
  The NaCl-H2O and KCl-H2O systems.
  Journal of Colloid and Interface Science 114, 409-415 (1986).
  Valid at 25°C and for solutions of molality up to ~13 mol/kg
"""

import numpy as np

from ....constants import charge_numbers
from ...general import SolutionFormula
from ..ionic import ionic_strength


def ln_gamma_extended_debye_huckel(m, T, solute, coeffs):
    """Mix of Hamer & Wu 1972 and Tang, Munkelwitz and Wang 1986.

    Used for NaCl and KCl
    """
    z1, z2 = charge_numbers[solute]
    A, B, C, D, E, beta = coeffs
    I = ionic_strength(solute, m=m)

    ln_gamma = - z1 * z2 * A * np.sqrt(I) / (1 + B * np.sqrt(I))
    ln_gamma += (beta * I) + (C * I**2) + (D * I**3) + (E * I**4)

    return ln_gamma


class ActivityCoefficient_Tang_Base(SolutionFormula):

    source ='Tang'

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'm'

    with_water_reference = False

    def calculate(self, m, T):
        log_g = ln_gamma_extended_debye_huckel(
            m=m,
            T=T,
            solute=self.solute,
            coeffs=self.coeffs.values(),
        )
        return 10**(log_g)


class ActivityCoefficient_NaCl_Tang_Base(ActivityCoefficient_Tang_Base):

    solute = 'NaCl'
    concentration_range = (1e-9, 14)

    coeffs = {
        'A': 0.5108,
        'B': 1.37,
        'C': 4.803e-3,
        'D': -2.736e-4,
        'E': 0,
        'beta': 2.796e-2,
    }


class ActivityCoefficient_KCl_Tang_Base(ActivityCoefficient_Tang_Base):

    solute = 'KCl'
    concentration_range = (1e-9, 13)

    coeffs = {
        'A': 0.5108,
        'B': 1.35,
        'C': 7.625e-3,
        'D': -7.892e-4,
        'E': 2.492e-5,
        'beta': -9.842e-3,
    }



"""Steiger formulas for solubility of solutions.

Sources
-------
- Steiger, M.,
  Crystal growth in porous materials—I:
  The crystallization pressure of large crystals.
  Journal of Crystal Growth 282, 455-469 (2005).
  Valid at 25°C and up to 13.5 mol/kg

- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

(some info of domain of validity of expressions in the following paper:)
Dorn, J. & Steiger, M. Measurement and Calculation of Solubilities in the
Ternary System NaCH 3 COO + NaCl + H 2 O from 278 K to 323 K.
J. Chem. Eng. Data 52, 1784-1790 (2007).)
"""

import numpy as np
from pynverse import inversefunc

from ....constants import dissociation_numbers
from ....format import make_array_method
from ...general import SaturatedSolutionFormula
from ..steiger import coeffs_steiger2008_activity
from ..steiger import coeffs_steiger2008_solubility
from ..pitzer import PitzerActivity


class Solubility_Steiger_Base(SaturatedSolutionFormula):

    source = 'Steiger 2008'

    concentration_unit = 'm'

    temperature_unit = 'K'
    temperature_range = (0 + 273.15, 50 + 273.15)

    def _solubility_product(self, T):
        ln_K = coeffs_steiger2008_solubility.ln_K(crystal=self.crystal, T=T)
        return np.exp(ln_K)

    def _solute_activity(self, m, T):
        """Incorporates water activity if hydrated crystal"""
        coeffs = coeffs_steiger2008_activity.coeffs(solute=self.solute, T=T)
        pitz = PitzerActivity(T=T, solute=self.solute, **coeffs)

        K_s = pitz.solute_activity(m=m)

        if self.crystal_hydration:
            nu_w = self.crystal_hydration
            a_w = pitz.water_activity(m=m)
            K_w = a_w ** nu_w
        else:
            K_w = 1

        return K_s * K_w


    @make_array_method
    def calculate(self, T):
        """Make array because the inversion needs to be made at each temperature."""

        def _solute_activity(m):
            return self._solute_activity(m, T)

        _solute_molality = inversefunc(_solute_activity, domain=[0, 8])

        K = self._solubility_product(T)
        m = _solute_molality(K)

        return m


# =============================== Steiger 2008 ===============================


class Solubility_NaCl_Steiger2008_Base(Solubility_Steiger_Base):
    crystal = 'NaCl'


class Solubility_NaCl_2H2O_Steiger2008_Base(Solubility_Steiger_Base):
    crystal = 'NaCl,2H2O'
    crystal_hydration = 2

    # Unsure of Tmin, but definitely its stability stops at 0.1°C
    temperature_range = (-20 + 273.15, 0.1 + 273.15)


class Solubility_Na2SO4_Steiger2008_Base(Solubility_Steiger_Base):
    """Thenardite"""
    crystal = 'Na2SO4'


class Solubility_Na2SO4_10H2O_Steiger2008_Base(Solubility_Steiger_Base):
    """Mirabilite"""
    crystal = 'Na2SO4,10H2O'
    crystal_hydration = 10

    # Mirabilite stable zone stops at 32.38°C
    temperature_range = (0 + 273.15, 32.38 + 273.15)


class Solubility_KCl_Steiger2008_Base(Solubility_Steiger_Base):
    crystal = 'KCl'



"""Gathers the formulas for the refractive index of KCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns one parameter:
- n (index of refraction, dimensionless)
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Tan (default) : "Dependence of Refractive Index on Concentration and
Temperature in Electrolyte Solution, Polar Solution, Nonpolar Solution, and
Protein Solution", Tan & Huang, J. Chem. Eng. Data  (2015).
Valid from w = 0 to w = 0.15 and for temperatures between 20 and 45°C
"""

from .tan import RefractiveIndex_KCl_Tan_Base


class RefractiveIndex_KCl_Tan(RefractiveIndex_KCl_Tan_Base):
    """Already defined in tan module"""
    default = True


# ============================= WRAP-UP FORMULAS =============================

RefractiveIndexFormulas_KCl = (
    RefractiveIndex_KCl_Tan,
)

# ====================== DIRECT RUN (test of formulas) =======================

if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import numpy as np

    from aquasol.solutions import refractive_index

    ws = 0.01, 0.05, 0.10, 0.15          # experimental weight fraction
    ww = np.linspace(0, 0.15, 100)       # weight fraction for fit

    nexps = {25: [1.3345, 1.3401, 1.3480, 1.3571], # 25°C
             45: [1.3320, 1.3379, 1.3460, 1.3550]} # 45°C

    fig, ax = plt.subplots()

    for T, ns in nexps.items():

        nn = refractive_index(w=ww, T=T, solute='KCl')

        ax.plot(ws, ns, '.')
        ax.plot(ww, nn, '-')

    plt.show()


"""Gathers the formulas for the activity coefficients of Na2SO4 solutions.

Sources
-------

- Steiger, M.,
  Crystal growth in porous materials—I:
  The crystallization pressure of large crystals.
  Journal of Crystal Growth 282, 455-469 (2005).
  Valid at 25°C and up to 13.5 mol/kg

- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

Note: some potential info about validity in:
Steiger, M. & Asmussen, S.
Crystallization of sodium sulfate phases in porous materials:
The phase diagram Na2SO4-H2O and the generation of stress.
Geochimica et Cosmochimica Acta 72, 4291-4306 (2008).

For now, I have assumed validity similar to NaCl (in temperatures)
"""

from .steiger import ActivityCoefficient_Na2SO4_Steiger2005_Base
from .steiger import ActivityCoefficient_Na2SO4_Steiger2008_Base


class ActivityCoefficient_Na2SO4_Steiger2005(ActivityCoefficient_Na2SO4_Steiger2005_Base):
    """Already defined in steiger module"""
    pass


class ActivityCoefficient_Na2SO4_Steiger2008(ActivityCoefficient_Na2SO4_Steiger2008_Base):
    """Already defined in steiger module"""
    default = True


# ========================== WRAP-UP OF FORMULAS =============================

ActivityCoefficientFormulas_Na2SO4 = (
    ActivityCoefficient_Na2SO4_Steiger2008,
    ActivityCoefficient_Na2SO4_Steiger2005,
)



"""Gathers the formulas for the activity of CaCl2 solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns one parameter
- a, water activity (dimensionless, range 0-1)
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Conde, M. R., Properties of aqueous solutions of lithium and calcium
chlorides: formulations for use in air conditioning equipment design.
International Journal of Thermal Sciences 43, 367-382 (2004).
"""


from .conde import WaterActivity_CaCl2_Conde_Base


class WaterActivity_CaCl2_Conde(WaterActivity_CaCl2_Conde_Base):
    """Already defined in conde module"""
    default = True


# ========================== WRAP-UP OF FORMULAS =============================

WaterActivityFormulas_CaCl2 = (
    WaterActivity_CaCl2_Conde,
)



"""Function to calculate the density of saturated water as a function of
temperature using IAPWS recommended equation and others.

Sources
-------

--- 'IAPWS'
    Wagner and Pruß
    "The IAPWS Formulation 1995 for the Thermodynamic Properties of Ordinary
    Water Substance for General and Scientific Use"
    (2002)
    Equation 2.6 page 399

--- 'Conde'
    Conde, M. R.
    Properties of aqueous solutions of lithium and calcium chlorides:
    formulations for use in air conditioning equipment design.
    International Journal of Thermal Sciences 43, 367-382
    (2004)
"""

from ...constants import Tc, rhoc
from ..general import WaterFormula


class DensitySat_IAPWS(WaterFormula):

    source ='IAPWS'
    temperature_unit = 'K'
    temperature_range = (273.15, Tc)
    default = True

    coeffs = [
        1.99274064,
        1.09965342,
        -0.510839303,
        -1.75493479,
        -45.5170352,
        -6.74694450e5,
    ]

    def calculate(self, T):
        """Saturated water density according to Wagner and Pruss 2002 (IAPWS 95)

        Input
        -----
        Temperature in K

        Output
        ------
        Density in kg/m^3

        Reference
        ---------
        Wagner and Pruß : "The IAPWS Formulation 1995 for the Thermodynamic Properties
        of Ordinary Water Substance for General and Scientific Use" (2002), eq. (2.6)

        Notes
        -----
        - Used by Al Ghafri 2012
        - Valid between triple point (0.01°C) and critical temperature 647.096K
        """
        c1, c2, c3, c4, c5, c6 = self.coeffs
        phi = 1 - T / Tc
        rho = rhoc * (1 + c1 * phi**(1/3) + c2 * phi**(2/3) + c3 * phi**(5/3)
                        + c4 * phi**(16/3) + c5 * phi**(43/3) + c6 * phi**(110/3))
        return rho


class DensitySat_Conde(WaterFormula):

    source ='Conde'
    temperature_unit = 'K'
    temperature_range = (273.15, Tc)

    coeffs = [
        1.9937718430,
        1.0985211604,
        -0.5094492996,
        -1.7619124270,
        -44.9005480267,
        -723692.2618632,
    ]

    def calculate(self, T):
        """Water density equation that looks very similar to Wagner, used by Conde.

        Input
        -----
        Temperature in C

        Output
        ------
        Density in kg/m^3

        Reference
        ---------
        Conde, M. R., Properties of aqueous solutions of lithium and calcium
        chlorides: formulations for use in air conditioning equipment design.
        International Journal of Thermal Sciences 43, 367-382 (2004).

        """
        c1, c2, c3, c4, c5, c6 = self.coeffs
        phi = 1 - T / Tc
        rho = rhoc * (1 + c1 * phi**(1/3) + c2 * phi**(2/3) + c3 * phi**(5/3)
                        + c4 * phi**(16/3) + c5 * phi**(43/3) + c6 * phi**(110/3))
        return rho


# ========================== WRAP-UP OF FORMULAS =============================

DensitySatFormulas = (
    DensitySat_IAPWS,
    DensitySat_Conde
)
